
from pymatgen.io.vasp.inputs import Poscar

from defermi.testing.core import DefermiTest
from defermi.generator import create_vacancies, create_substitutions


class TestDefectGenerator(DefermiTest):

    @classmethod
    def setUpClass(cls):
        self = cls()
        structure = Poscar.from_file(self.get_testfile_path('BaTiO3_mp-5986.poscar')).structure
        cls.structure = structure

    
    def test_create_vacancies(self):
        defects = create_vacancies(
                                structure=self.structure,
                                elements=None,
                                supercell_size=3)
        
        self.assertEqual( len(defects), 4)
        
        def count_species(defects,element):
            count = 0
            for df in defects:
                if df.specie == element:
                    count += 1
            return count
        
        self.assertEqual(count_species(defects,'Ba'), 1)
        self.assertEqual(count_species(defects,'Ti'), 1)
        self.assertEqual(count_species(defects,'O'), 2)



    def test_create_substitutions(self):
        defects = create_substitutions(
                                structure=self.structure,
                                elements_to_replace={'Ba':'Sr','Ti':'Zr','O':'N'},
                                supercell_size=3)
        
        self.assertEqual( len(defects), 4)
        
        def count_species(defects,element):
            count = 0
            for df in defects:
                if df.specie == element:
                    count += 1
            return count
        
        self.assertEqual(count_species(defects,'Sr'), 1)
        self.assertEqual(count_species(defects,'Zr'), 1)
        self.assertEqual(count_species(defects,'N'), 2)



import argparse


def main():
    """
    Handle main.
    """    
    parser = argparse.ArgumentParser(
        description="""
        Command line interface for defermi. Type "defermi -h" to view options.
        """
        )

    subparsers = parser.add_subparsers()

    # add setup functions here

    args = parser.parse_args()
    
    try:
        args.func
    except AttributeError:
        parser.print_help()
        raise SystemExit("Please specify a command.")
    return args.func(args)


if __name__ == '__main__':
    main()



from pymatgen.core.sites import PeriodicSite
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from .defects import Interstitial, Substitution, Vacancy
from .tools.structure import remove_oxidation_state_from_site



def create_interstitials(structure,elements,supercell_size=None,**kwargs):
    """
    Create Interstitial objects based on Voronoi with pymatgen,
    staring from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure.
    elements : list
        List of element symbols.
    supercell_size : int
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. The default is None.
    kwargs: 

        Kwargs to pass to VoronoiInterstitialGenerator:

        - clustering_tol: Tolerance for clustering the Voronoi nodes.
        - min_dist: Minimum distance between an interstitial and the nearest atom.
        - ltol: Tolerance for lattice matching.
        - stol: Tolerance for structure matching.
        - angle_tol: Angle tolerance for structure matching.
        - kwargs: Additional keyword arguments for the `TopographyAnalyzer` constructor.


    Returns
    -------
    defects : list
        List of Interstitial objects.

    """
    from pymatgen.analysis.defects.generators import VoronoiInterstitialGenerator
    
    defects = []
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    generator = VoronoiInterstitialGenerator().generate(bulk_structure,elements)
    for inter in generator:
        bulk_structure.remove_oxidation_states()
        remove_oxidation_state_from_site(inter.site)
        interstitial = Interstitial(
                                defect_site=inter.site,
                                bulk_structure=bulk_structure,
                                multiplicity=inter.multiplicity,
                                label=f'mult{inter.multiplicity}')
        defects.append(interstitial)
    return defects


def create_substitutions(structure,elements_to_replace,supercell_size=None, **kwargs):
    """
    Create Substitution objects for each non-equivalent site,
    staring from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements_to_replace : str
        Dict with element symbol of specie to be replaced as keys and element 
        symbol of the species to be replaced with as values ({'old_El':'new_El'}).
    supercell_size : int or numpy array
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. 
    kwargs : dict
    
        Kwargs to pass to SpaceGroupAnalyzer class:

        - symprec (float): 
            Tolerance for symmetry finding. Defaults to 0.01,
            which is fairly strict and works well for properly refined
            structures with atoms in the proper symmetry coordinates. For
            structures with slight deviations from their proper atomic
            positions (e.g., structures relaxed with electronic structure
            codes), a looser tolerance of 0.1 (the value used in Materials
            Project) is often needed.

        - angle_tolerance (float): 
            Angle tolerance for symmetry finding. Defaults to 5 degrees.


    Returns
    -------
    defects : list
        List of Substitution objects.

    """
    import string
    letters = string.ascii_lowercase

    defects = [] 
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    else:
        sites = bulk_structure.sites

    sym_struct = SpacegroupAnalyzer(structure=bulk_structure,**kwargs).get_symmetrized_structure()
    for el_to_sub,el_subbed in elements_to_replace.items():
        idx = 0
        for site_group in sym_struct.equivalent_sites:
            site = site_group[0]
            if site.specie.symbol == el_to_sub:
                defect_site = PeriodicSite(el_subbed,site.frac_coords,site.lattice)
                substitution = Substitution(
                                        defect_site=defect_site,
                                        bulk_structure=bulk_structure,
                                        site_in_bulk=site)
                multiplicity = len(sym_struct.find_equivalent_sites(site))
                substitution.set_multiplicity(multiplicity)
                substitution.set_label(letters[idx])
                idx += 1
                defects.append(substitution)
    return defects   


def create_vacancies(structure,elements=None,supercell_size=None, **kwargs):
    """
    Create Vacancy objects for each non-equivalent site,
    staring from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements : str
        Symbol of the elements for which vacancies are needed.
        If None all of the elements are considered. The default is None.
    supercell_size : int or numpy array
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. 
    kwargs : dict
    
        Kwargs to pass to SpaceGroupAnalyzer class:

        - symprec (float): 
            Tolerance for symmetry finding. Defaults to 0.01,
            which is fairly strict and works well for properly refined
            structures with atoms in the proper symmetry coordinates. For
            structures with slight deviations from their proper atomic
            positions (e.g., structures relaxed with electronic structure
            codes), a looser tolerance of 0.1 (the value used in Materials
            Project) is often needed.

        - angle_tolerance (float): 
            Angle tolerance for symmetry finding. Defaults to 5 degrees.


    Returns
    -------
    defects : list
        List of Vacancy objects.
        
    """
    import string
    letters = string.ascii_lowercase

    defects = []
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    if not elements:
        elements = [el.symbol for el in bulk_structure.composition.elements]
        
    sym_struct = SpacegroupAnalyzer(structure=bulk_structure,**kwargs).get_symmetrized_structure()
    for el in bulk_structure.composition.elements:
        idx = 0
        for site_group in sym_struct.equivalent_sites:
            site = site_group[0]
            if el.symbol in elements:
                if site.specie == el:
                    vacancy = Vacancy(
                                    defect_site=site,
                                    bulk_structure=bulk_structure)
                    multiplicity = len(sym_struct.find_equivalent_sites(site))
                    vacancy.set_multiplicity(multiplicity)
                    vacancy.set_label(letters[idx])
                    idx += 1
                    defects.append(vacancy)
    return defects     


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 12 16:47:57 2023

@author: villa
"""
from pymatgen.core.sites import PeriodicSite
from pymatgen.core.structure import Structure

from defermi.tools.utils import get_object_from_json
from defermi.chempots.core import Chempots
from defermi.defects import Interstitial, Vacancy
from defermi.entries import DefectEntry

from defermi.testing.core import DefermiTest


class TestDefectEntry(DefermiTest):

    def setUp(self):
        mu_B = -6.6794
        mu_P = -5.4133
        mu_Si = -5.4224
        self.chempots = Chempots({'Si':mu_Si,'P':mu_P,'B':mu_B})
        self.vbm = 5.8268
        self.band_gap = 0.5729
        self.stress_bulk = [[19.5106052, -0.0, -0.0], [0.0, 19.51060527, 0.0], [0.0, 0.0, 19.51060528]]
        self.bulk_modulus = 144
        
    @property
    def entry(self):
        return self.get_entry()
    
    def get_entry(self):
        bulk_structure = get_object_from_json(Structure,self.get_testfile_path('Si-bulk_structure_3x3x3_supercell.json'))
        defect_site = PeriodicSite('Si',[0.5417, 0.5417, 0.5417], bulk_structure.lattice)
        defect = Interstitial(
                            specie = 'Si',
                            defect_site=defect_site,
                            bulk_structure=bulk_structure,
                            charge=-1.0,
                            multiplicity=54,
                            label='mult54')
        
        energy_diff = 5.168792819999965
        corrections = {'kumagai': -0.08825262621975172}
        data = {'stress': 
                [[39.5772376, -0.0, 0.0],
                 [0.0, 39.57723773, 0.0],
                 [0.0, -0.0, 39.57723775]]}
        entry = DefectEntry(defect,energy_diff,corrections,data)
        return entry 
    
    
    def test_formation_energy(self):
        formation_energy = self.entry.formation_energy(self.vbm,self.chempots,fermi_level=0.25)
        self.assert_all_close(formation_energy,4.423740193780213,rtol=100,atol=1e-02)
        
    def test_defect_concentration(self):
        concentration = self.entry.defect_concentration(self.vbm,self.chempots,temperature=300,fermi_level=0.25) 
        self.assert_all_close(concentration,2.4154948187001994e-52,rtol=100,atol=1e-02)
        
        def_conc = self.entry.defect_concentration(self.vbm,self.chempots,temperature=300,fermi_level=5) 
        self.assert_all_close(def_conc,4.995386515296238e+22,rtol=100,atol=1e-02)
        
    def test_relaxation_volume(self):
        relaxation_volume = self.entry.relaxation_volume(self.stress_bulk,self.bulk_modulus) # doesn't make physical sense with charged defect
        self.assert_all_close(relaxation_volume,-15.063823947648379)

    def test_label(self):
        entry = self.entry
        label = 'test'
        entry.set_label(label)
        actual = entry.label
        desired = label
        self.assertEqual(actual, desired)
        
    def test_as_dict_from_dict(self):
        entry_dict_1 = self.entry.as_dict()
        entry_dict_2 = DefectEntry.from_dict(self.entry.as_dict()).as_dict()
        self.assert_object_almost_equal(entry_dict_1,entry_dict_2)


    def test_custom_functions(self):
        entry = DefectEntry(Vacancy('O',charge=2,bulk_volume=800,multiplicity=1),energy_diff=7)
        chempots = {'O': -4.95}
        vbm = 0
        band_gap = 2

        def custom_eform(entry,vbm=None,chemical_potentials=None,fermi_level=0,temperature=300,**kwargs):
                formation_energy = entry.energy_diff + entry.charge*(vbm+fermi_level) 
                if chemical_potentials:
                    chempot_correction = -1 * sum([entry.delta_atoms[el]*chemical_potentials[el] for el in entry.delta_atoms])
                else:
                    chempot_correction = 0
                    
                formation_energy = formation_energy + chempot_correction
                test_quantity = kwargs['test'] if kwargs and 'test' in kwargs.keys() else 0
                return formation_energy - 1/500 *temperature + test_quantity       
        

        def custom_dconc(entry,vbm=0,chemical_potentials=None,temperature=0,fermi_level=0,per_unit_volume=True,eform_kwargs={},**kwargs):
                from defermi.entries import fermi_dirac
                n = entry.defect.site_concentration_in_cm3 if per_unit_volume else entry.multiplicity 
                if per_unit_volume:
                    n = n * (1+1/500*temperature)
                eform = entry.formation_energy(
                                        vbm=vbm,
                                        chemical_potentials=chemical_potentials,
                                        fermi_level=fermi_level,
                                        temperature=temperature,
                                        **eform_kwargs)
                
                test_quantity = kwargs['test_conc'] if kwargs and 'test_conc' in kwargs.keys() else 1
                return  n * fermi_dirac(eform,temperature) * test_quantity
        
        actual = entry.formation_energy(vbm=vbm,chemical_potentials=chempots,fermi_level=1)
        desired = 4.05
        self.assert_all_close(actual, desired)
        actual = entry.formation_energy(vbm=vbm,chemical_potentials=chempots,fermi_level=1,temperature=500)
        self.assert_all_close(actual, desired)

        entry.set_formation_energy_function(custom_eform)
        actual = entry.formation_energy(vbm=vbm,chemical_potentials=chempots,fermi_level=1,temperature=0)
        self.assert_all_close(actual, desired)
        actual = entry.formation_energy(vbm=vbm,chemical_potentials=chempots,fermi_level=1,temperature=500)
        desired = 3.05
        self.assert_all_close(actual, desired)

        actual = entry.formation_energy(vbm=vbm,chemical_potentials=chempots,fermi_level=1,temperature=500,test=2)
        desired = 5.05
        self.assert_all_close(actual, desired)

        actual = entry.formation_energy(vbm=vbm,chemical_potentials=chempots,fermi_level=1,test=2)
        desired = 6.05
        self.assert_all_close(actual, desired)


        actual = entry.defect_concentration(vbm=0,chemical_potentials=chempots,fermi_level=0,temperature=800)
        desired = 1.8255765558702477e+18
        self.assert_all_close(actual, desired, rtol=1e-03)

        entry.set_defect_concentration_function(custom_dconc)
        actual = entry.defect_concentration(vbm=0,chemical_potentials=chempots,fermi_level=0,temperature=800)
        desired = 4.746499045262644e+18
        self.assert_all_close(actual, desired, rtol=1e-03)

        actual = entry.defect_concentration(vbm=0,chemical_potentials=chempots,fermi_level=0,temperature=800,eform_kwargs={'test':1})
        desired = 2383885542239.6626
        self.assert_all_close(actual, desired, rtol=1e-03)

        actual = entry.defect_concentration(vbm=0,chemical_potentials=chempots,fermi_level=0,temperature=800,eform_kwargs={'test':1},test_conc=10)
        desired = desired * 10
        self.assert_all_close(actual, desired, rtol=1e-03)

        actual = entry.defect_concentration(vbm=0,chemical_potentials=chempots,fermi_level=0,temperature=800,test_conc=10)
        desired = 4.746499045262644e+19
        self.assert_all_close(actual, desired, rtol=1e-03)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr  3 14:08:42 2023

@author: villa
"""
import importlib
import warnings
from abc import ABCMeta, abstractmethod, abstractproperty
from monty.json import MSONable
import importlib

from pymatgen.core.composition import Composition
from pymatgen.core.sites import PeriodicSite
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from .tools.structure import is_site_in_structure, is_site_in_structure_coords



class Defect(MSONable,metaclass=ABCMeta): #MSONable contains as_dict and from_dict methods
    """
    Abstract class for a single point defect
    """

    def __init__(self, 
                specie=None, 
                defect_site=None,
                bulk_structure=None,
                charge=None,
                multiplicity=None,
                bulk_volume=None,
                label=None): 
        """
        Base class for defect objets.

        Parameters
        ----------
        specie : str
            Defect element symbol.
        defect_site : Site
            Pymatgen Site object of the defect.
        bulk_structure : Structure
            Pymatgen Structure without defects.
        charge : int or float
            Charge of the defect.
        multiplicity : int
            Multiplicity of defect within the simulation cell.
        bulk_volume : float
            Volume of bulk cell in A°^3.
        label : str
            Defect label.

        """
        if  specie:
            self._specie = specie
        elif defect_site:
            self._specie = defect_site.specie.symbol
        else:
            raise ValueError('Either defect species symbol or defect Site have to be provided')
        self._defect_site = defect_site
        self._bulk_structure = bulk_structure
        self._charge = charge
        self._multiplicity = multiplicity
        self._bulk_volume = bulk_volume
        self._label = label


    def __repr__(self):
        string = f'Defect: type={self.type}, species={self.specie}'
        if self.charge is not None:
            string += f', charge={self.charge}'
        if self.label:
            string += f', label={self.label}'
        if self._defect_site:
            string +=  ", site= {}".format(self.site.frac_coords)
        return string
    
    def __print__(self):
        return self.__repr__()
    
    def __iter__(self):
        """
        Dummy iterable for integration with defect complexes
        """
        return [self].__iter__()

    @staticmethod
    def from_string(string, **kwargs):
        if '(' in string:
            name,label = string.split('(')
            label = label.strip(')')
        else:
            name = string
            label=None
        nsplit = name.split('_')
        ntype = nsplit[0]
        el = nsplit[1]
        bulk_specie = None
        if ntype=='Vac':
            dtype = 'Vacancy'
            dspecie = el
        elif ntype=='Int':
            dtype = 'Interstitial'
            dspecie = el
        elif ntype=='Sub':
            dtype = 'Substitution'
            dspecie = el
            el_bulk = nsplit[3]
            bulk_specie = el_bulk
        elif ntype=='Pol':
            dtype = 'Polaron'
            dspecie = el

        module = importlib.import_module(Defect.__module__)
        defect_class = getattr(module,dtype)
        kwargs.update({
            'specie':dspecie,
            'label': label
            })
        if dtype=='Substitution':
            kwargs['bulk_specie'] = bulk_specie

        return defect_class(**kwargs)
            

    @property
    def bulk_structure(self):
        """
        Structure without defects.
        """
        if self._bulk_structure:
            return self._bulk_structure
            #warnings.warn('Bulk structure is not stored in Defect object')

    @property
    def bulk_volume(self):
        """
        Volume of bulk cell in A°^3.
        """
        if self._bulk_volume:
            return self._bulk_volume
        elif self.bulk_structure:
            return self.bulk_structure.volume
        else:
            warnings.warn('Neither bulk structure nor the bulk cell volume were assigned')
    
    @property
    def charge(self):
        """
        Charge of the defect.
        """
        if self._charge is None:
            None #warnings.warn('Charge was not assigned to defect object. Use set_charge to assign it.')
        else:
            return self._charge
     
    @abstractproperty
    def defect_composition(self):
        """
        Defect composition as a Composition object
        """
        return

    @abstractproperty
    def defect_site_index(self):
        """
        Index of the defect site in the structure
        """
        return
    
    @property
    def defects(self): # dummy property to integrate Defect and DefectComplex
        return [self]
    
    @property
    def specie(self):
        """
        Defect species.
        """
        return self._specie
    
    @property
    def defect_structure(self):
        """
        Structure of the defect.
        """
        return self.generate_defect_structure() 
    
    @property
    def type(self):
        """
        Defect type.
        """
        return self.__class__.__name__
    
    @property
    def delta_atoms(self):
        """
        Dictionary with defect element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return

    @abstractmethod
    def get_multiplicity(self):
        return 

    @property
    def label(self):
        """
        Defect label.
        """
        return self._label

    @property
    def multiplicity(self):
        """
        Multiplicity of a defect site within the structure
        """
        return self._multiplicity
      
    @abstractproperty  
    def name(self):
        """
        Name of the defect.
        """
        return

    @property
    def site(self):
        """
        Defect position as a Site object
        """
        if self._defect_site:
            return self._defect_site
        else:
            warnings.warn('Site is not stored in Defect object')
   
    @property
    def site_concentration_in_cm3(self):
        """
        Site concentration (multiplicity/volume) expressed in cm^-3.
        """
        return self.multiplicity * 1e24 / self.bulk_volume 
   
    @property  
    def symbol(self):
        """
        Latex formatted name of the defect.
        """
        return 

    @property
    def symbol_with_charge(self):
        """
        Name in latex format with charge written as a number.
        """
        return format_legend_with_charge_number(self.symbol,self.charge)
    
    @property
    def symbol_with_charge_kv(self):
        """
        Name in latex format with charge written with Kröger and Vink notation.
        """
        return format_legend_with_charge_kv(self.symbol,self.charge)
    
    def set_bulk_volume(self, new_volume):
        """
        Sets the volume of bulk cell
        """
        self._bulk_volume = new_volume
        return

    def set_charge(self, new_charge=0.0):
        """
        Sets the charge of the defect.
        """
        self._charge = new_charge
        return
    
    def set_label(self,new_label):
        """
        Sets the label of the defect
        """
        self._label = new_label
        return

    def set_multiplicity(self, new_multiplicity=None):
        """
        Sets the Multiplicity of the defect.
        If `new_multiplicity` is not provided it is determined 
        with `self.get_multiplicity`
        """
        if new_multiplicity is None:
            new_multiplicity = self.get_multiplicity()
        self._multiplicity = new_multiplicity
        return
        
        
class Vacancy(Defect):
    
    """
    Subclass of Defect for single vacancies.
    """
    @property
    def defect_composition(self):
        """
        Composition of the defect.
        """
        temp_comp = self.bulk_structure.composition.as_dict()
        temp_comp[str(self.site.specie)] -= 1
        return Composition(temp_comp)
    
    @property
    def defect_site_index(self):
        """
        Index of the defect site in the bulk structure
        """
        _,index = is_site_in_structure_coords(self.site, self.bulk_structure)
        return index
    
    @property
    def delta_atoms(self):
        """
        Dictionary with element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {self.specie:-1}
    
    def generate_defect_structure(self,bulk_structure=None,defect_site_index=None):
        """
        Generate a structure containing the defect starting from a bulk structure.

        Parameters
        ----------
        bulk_structure : Structure
            Bulk Structure. If not provided `self.bulk_structure` is used.
        defect_site_index : Structure
            Index of the defect site in the bulk structure. If not provided 
            `self.defect_site_index` is used.

        Returns
        -------
        structure : Structure
            Structure containing the defect.

        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        defect_site_index = defect_site_index if defect_site_index else self.defect_site_index
        defect_structure = bulk_structure.copy()
        defect_structure.remove_sites([defect_site_index])
        return defect_structure
    
    def get_multiplicity(self,**kwargs):
        """
        Get multiplicity of the defect in the structure.

        Parameters
        ----------
        **kwargs : dict
            Kwargs to pass to `SpacegroupAnalyzer` ("symprec", "angle_tolerance")
        """
        sga = SpacegroupAnalyzer(self.bulk_structure,**kwargs)
        symmetrized_structure = sga.get_symmetrized_structure()
        equivalent_sites = symmetrized_structure.find_equivalent_sites(self.site)
        return len(equivalent_sites)

    @property
    def name(self):
        """
        Name of the defect.
        """
        name = f"Vac_{self.specie}"
        if self.label:
            name += f'({self.label})'
        return name
    
    @property
    def symbol(self):
        symbol = "$V_{%s}$" %self.specie
        if self.label:
            symbol += '(%s)' %self.label
        return symbol 
        
        
    
class Substitution(Defect):
    """
    Subclass of Defect for substitutional defects.
    """

    def __init__(self,
                specie=None,
                bulk_specie=None,
                defect_site=None,
                bulk_structure=None,
                charge=None,
                multiplicity=None,
                bulk_volume=None,
                label=None,
                site_in_bulk=None):
        """
        Parameters
        ----------
        site_in_bulk : PeriodicSite
            Original Site in bulk structure were substitution took place.

        """
        super().__init__(specie=specie,
                        defect_site=defect_site,
                        bulk_structure=bulk_structure,
                        charge=charge,
                        multiplicity=multiplicity,
                        bulk_volume=bulk_volume,
                        label=label)
        if bulk_specie:
            self._bulk_specie = bulk_specie 
        elif site_in_bulk:
            self._bulk_specie = site_in_bulk.specie.symbol
        else:
            raise ValueError('Either bulk species or substitution site in bulk structure have to be provided')
        
        self._site_in_bulk = site_in_bulk 


    @property
    def bulk_specie(self):
        return self._bulk_specie

    @property  
    def defect_composition(self):
        """
        Composition of the defect.
        """
        defect_index = self.defect_site_index
        comp = self.bulk_structure.composition.as_dict()
        if str(self.specie) not in comp.keys():
            comp[str(self.specie)] = 0
        comp[str(self.specie)] += 1
        comp[str(self.bulk_structure[defect_index].specie)] -= 1
        return Composition(comp)

    @property
    def defect_site_index(self):
        """
        Index of the defect site in the bulk structure.
        """
        return self.bulk_structure.index(self.site_in_bulk)

    @property
    def delta_atoms(self):
        """
        Dictionary with element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {self.specie:1, self.bulk_specie:-1}
    
    def generate_defect_structure(self,bulk_structure=None,defect_site_index=None):
        """
        Generate a structure containing the defect starting from a bulk structure.

        Parameters
        ----------
        bulk_structure : Structure
            Bulk Structure. If not provided `self.bulk_structure` is used.
        defect_site_index : Structure
            Index of the defect site in the bulk structure. If not provided 
            `self.defect_site_index` is used.

        Returns
        -------
        structure : Structure
            Structure containing the defect.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        defect_site_index = defect_site_index if defect_site_index else self.defect_site_index
        defect_structure = bulk_structure.copy()
        defect_structure.replace(defect_site_index,self.specie)  
        return defect_structure

    def get_multiplicity(self,**kwargs):
        """
        Get multiplicity of the defect in the structure

        Parameters
        ----------
        **kwargs : dict
            Kwargs to pass to `SpacegroupAnalyzer` ("symprec", "angle_tolerance")
        """
        sga = SpacegroupAnalyzer(self.bulk_structure,**kwargs)
        symmetrized_structure = sga.get_symmetrized_structure()
        equivalent_sites = symmetrized_structure.find_equivalent_sites(self.site_in_bulk)
        return len(equivalent_sites)

    def get_site_in_bulk(self):
        if self.bulk_structure and self.site:
            try:
                site = min(
                    self.bulk_structure.get_sites_in_sphere(self.site.coords, 0.5, include_index=True),
                            key=lambda x: x[1])  
                # there's a bug in pymatgen PeriodicNeighbour.from_dict and the specie attribute, get PeriodicSite instead
                site = PeriodicSite(site.species, site.frac_coords, site.lattice)
                return site
            except:
                return ValueError("""No equivalent site has been found in bulk, defect and bulk structures are too different.\
    Try using the unrelaxed defect structure or provide bulk site manually""")
        else:
            return None

    @property 
    def name(self):
        """
        Name for this defect.
        """
        name = f'Sub_{self.specie}_on_{self.bulk_specie}'
        if self.label:
            name += f'({self.label})'
        return name   
    
    @property
    def symbol(self):
        symbol = "$%s_{%s}$" %(self.specie, self.bulk_specie)   
        if self.label:
            symbol += '(%s)' %self.label
        return symbol 
    
    @property
    def site_in_bulk(self):
        if self._site_in_bulk:
            return self._site_in_bulk
        else:
            return self.get_site_in_bulk()


    
class Interstitial(Defect):
    """
    Subclass of Defect for interstitial defects.
    """
    
    @property
    def defect_composition(self):
        """
        Composition of the defect.
        """
        temp_comp = self.bulk_structure.composition.as_dict()
        temp_comp[str(self.site.specie)] += 1
        return Composition(temp_comp)
    
    @property
    def defect_site_index(self):
        """
        Index of the defect site in the defect structure
        """
        is_site,index = is_site_in_structure(self.site, self.defect_structure)
        return index  # more flexibility with is_site_in_structure
       
    @property
    def delta_atoms(self):
        """
        Dictionary with element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {self.specie:1}
    
    def generate_defect_structure(self,bulk_structure=None):
        """
        Generate a structure containing the defect starting from a bulk structure.

        Parameters
        ----------
        bulk_structure : Structure
            Bulk Structure. If not provided `self.bulk_structure` is used.

        Returns
        -------
        structure : Structure
            Structure containing the defect.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        defect_structure = bulk_structure.copy()
        defect_structure.append(self.site.species,self.site.frac_coords)
        return defect_structure
        

    def get_multiplicity(self):
        raise NotImplementedError('Multiplicity calculation not implemented for Interstitial')

    @property
    def name(self):
        """
        Name of the defect.
        """
        name = f"Int_{self.specie}"
        if self.label:
            name += f'({self.label})'
        return name

    @property
    def symbol(self):
        symbol = "$%s_i$" %(self.specie)     
        if self.label:
            symbol += '(%s)' %self.label
        return symbol    
        
      
class Polaron(Defect):
    """
    Subclass of Defect for polarons.
    """   
    
    def __init__(self,
                specie=None,
                defect_site=None,
                bulk_structure=None,
                charge=None,
                multiplicity=None,
                label=None,
                bulk_volume=None,
                defect_structure=None):
        """
        defect_structure: Structure
            Structure containing the polaron. If not provided the site index is searched 
            in the bulk structure, and `defect_structure` is set equal to the bulk structure.
        """
        super().__init__(specie=specie,
                        defect_site=defect_site,
                        bulk_structure=bulk_structure,
                        charge=charge,
                        multiplicity=multiplicity,
                        bulk_volume=bulk_volume,
                        label=label)
        self._defect_structure = defect_structure
        
        
    @property
    def defect_composition(self):
        """
        Composition of the defect.
        """
        return self.bulk_structure.composition
    
    @property
    def defect_site_index(self):
        """
        Index of the defect site in the structure.
        """
        return self.defect_structure.index(self.site)
        
    @property
    def delta_atoms(self):
        """
        Dictionary with delement as keys and difference in particle number 
        between defect and bulk structure as values.
        """
        return {}
    
    def generate_defect_structure(self,bulk_structure=None):
        """
        Structure containing the polaron. If not provided the site index is searched 
        in the bulk structure, and `defect_structure` is set equal to the bulk structure.
        """
        if self._defect_structure:
            return self._defect_structure
        else:
            bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
            return bulk_structure  
        
        
    def get_multiplicity(self,**kwargs):
        """
        Get multiplicity of the defect in the structure.

        Parameters
        ----------
        **kwargs : dict
            Kwargs to pass to `SpacegroupAnalyzer` ("symprec", "angle_tolerance")
        """
        sga = SpacegroupAnalyzer(self.bulk_structure,**kwargs)
        symmetrized_structure = sga.get_symmetrized_structure()
        equivalent_sites = symmetrized_structure.find_equivalent_sites(self.site)
        return len(equivalent_sites)

    @property
    def name(self):
        """
        Name of the defect.
        """
        name = f"Pol_{self.specie}"
        if self.label:
            name += f'({self.label})'
        return name
    
    @property
    def symbol(self):
        symbol = "$%s_{%s}$" %(self.specie,self.specie)
        if self.label:
            symbol += '(%s)' %self.label
        return symbol   
    
    
class DefectComplex(MSONable,metaclass=ABCMeta):

    def __init__(self,
                defects,
                bulk_structure=None,
                charge=None,
                multiplicity=None,
                bulk_volume=None,
                label=None):
        """
        Class to describe defect complexes

        Parameters
        ----------
        defects : list
            List of Defect objects.
        bulk_structure : Structure
            Pymatgen Structure of the bulk material.
        charge : int or float
            Charge of the defect.
        multiplicity : int
            Multiplicity of the defect.
        """
        self._defects = defects
        self._bulk_structure = bulk_structure
        self._charge = charge
        self._multiplicity = multiplicity
        self._bulk_volume = bulk_volume
        self._label = label


    def __repr__(self):
        string = 'DefectComplex: ['
        for df in self.defects:
            string += f' {df.__repr__()},'
        string += ' ] '
        return string
    
    def __str__(self):
        return self.__repr__()
    
    def __iter__(self):
        return self.defects.__iter__()
    
    @staticmethod
    def from_string(string, **kwargs):
        """
        Get `DefectComplex` object from string.
        """
        names = string.split(';')
        defects = [Defect.from_string(n) for n in names]
        return DefectComplex(defects=defects, **kwargs)
    

    @property
    def bulk_structure(self):
        """
        Structure without defects.
        """
        if self._bulk_structure:
            return self._bulk_structure
        else:
            warnings.warn('Bulk structure is not stored in Defect object')

    @property
    def bulk_volume(self):
        """
        Volume of bulk cell in A°^3.
        """
        if self._bulk_volume:
            return self._bulk_volume
        elif self.bulk_structure:
            return self.bulk_structure.volume
        else:
            warnings.warn('Neither bulk structure nor the bulk cell volume were assigned')
    
    @property
    def charge(self):
        """
        Charge of the defect.
        """
        if self._charge is None:
            warnings.warn('Charge was not assigned to defect object. Use set_charge to assign it.')
        else:
            return self._charge

    @property
    def defects(self):
        """
        List of single defects consituting the complex.
        """
        return self._defects

    @property
    def defect_composition(self):
        """
        Composition of defect structure.
        """
        return self.defect_structure.composition
        
    @property
    def defect_names(self):
        """
        List of names of the single defects.
        """
        return [d.name for d in self.defects]

    @property
    def defect_structure(self):
        """
        Structure containing the defect.
        """
        return self.generate_defect_structure()

    @property
    def type(self):
        """
        Defect type.
        """
        return self.__class__.__name__
    
    @property
    def delta_atoms(self):
        """
        Dictionary with Element as keys and particle difference between defect structure
        and bulk structure as values.
        """
        da_global = None
        for d in self.defects:
            da_single = d.delta_atoms
            if da_global is None:
                da_global = da_single.copy()
            else:
                for e in da_single:
                    prec = da_global[e] if e in da_global.keys() else 0
                    da_global[e] = prec + da_single[e]       
        return da_global 

    def generate_defect_structure(self,bulk_structure=None):
        """
        Generate a structure containing the defect starting from a bulk structure.
        If not provided `self.bulk_structure` is used.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        structure = bulk_structure.copy()
        for df in self.defects:
            df_structure = df.generate_defect_structure(structure)
            structure = df_structure.copy()
        return structure

    def get_multiplicity(self):
        raise NotImplementedError('Not implemented for DefectComplex')

    @property
    def label(self):
        """
        Defect label.
        """
        return self._label

    @property
    def multiplicity(self):
        """
        Multiplicity of a defect site within the structure
        """
        return self._multiplicity

    @property
    def name(self):
        """
        Name of the defect. Behaves like a string with additional attributes.
        """
        name = ';'.join([df.name for df in self.defects])
        if self.label:
            name += f'({self.label})'
        return name
    
    @property
    def site_concentration_in_cm3(self):
        """
        Site concentration (multiplicity/volume) expressed in cm^-3.
        """
        return self.multiplicity * 1e24 / self.bulk_volume 
    
    @property
    def sites(self):
        """
        Site objects of single defects.
        """
        return [d.site for d in self.defects]
    
    @property
    def symbol(self):
        """
        Latex formatted name of the defect.
        """
        symbol = '-'.join([df.symbol for df in self.defects])  #'-'.join([df.symbol.split('(')[0] for df in self.defects]) without single df labels
        if self.label:
            symbol = symbol + '(%s)'%self.label 
        return symbol

    @property
    def symbol_with_charge(self):
        """
        Name in latex format with charge written as a number.
        """
        return format_legend_with_charge_number(self.symbol,self.charge)
    
    @property
    def symbol_with_charge_kv(self):
        """
        Name in latex format with charge written with kroger and vink notation.
        """
        return format_legend_with_charge_kv(self.symbol,self.charge)


    def set_charge(self, new_charge):
        """
        Sets the charge of the defect.
        """
        self._charge = new_charge
        return    

    def set_label(self,new_label):
        """
        Sets the label of the defect
        """
        self._label = new_label
        return

    def set_multiplicity(self, new_multiplicity):
        """
        Sets the charge of the defect.
        """
        self._multiplicity = new_multiplicity
        return
    


def get_defect_from_string(string, **kwargs):
    """
    Get defect object from a string (`Defect` or `DefectComplex`).
    """
    if ';' in string:
        return DefectComplex.from_string(string, **kwargs)
    else:
        return Defect.from_string(string, **kwargs)
   

def format_legend_with_charge_number(label,charge):
    """
    Get label in latex format with charge written as a number.
   
    Parameters
    ----------
    label : str
        Original name of the defect.
    charge : int or float
        Charge of the defect. Floats are converted to integers.
    
    Returns
    -------
    string : str
        Formatted defect name with charge.
    """
    s = label
    charge = int(charge)
    if charge > 0:
        q = '+'+str(charge)
    elif charge == 0:
        q = '\\;' + str(charge)
    else:
        q = str(charge)
    return s + '$^{' + q + '}$'


def format_legend_with_charge_kv(label,charge):
    """
    Get label in latex format with charge written with Kröger and Vink notation.

    Parameters
    ----------
    label : str
        Original name of the defect.
    charge : int or float
        Charge of the defect. Floats are converted to integer.

    Returns
    -------
    string : str
        Formatted defect name with Kröger and Vink notation.
    """
    mod_label = label + '$'
    charge = int(charge)
    if charge < 0:
        for i in range(0,abs(charge)):
            if i == 0:
                mod_label = mod_label + "^{"
            mod_label = mod_label + "'"
        mod_label = mod_label + "}"
    elif charge == 0:
        mod_label = mod_label + "^{x}"
    elif charge > 0:
        for i in range(0,charge):
            if i == 0:
                mod_label = mod_label + "^{"
            mod_label = mod_label + "°"
        mod_label = mod_label + "}"
    
    mod_label = mod_label + "$"
    
    return mod_label

    
def get_delta_atoms(structure_defect,structure_bulk):
    """
    Build delta_atoms dictionary starting from Pymatgen Structure objects.

    Parameters
    ----------
    structure_defect : Structure
        Defect structure.
    structure_bulk : Structure
        Bulk structure.

    Returns
    -------
    delta_atoms : dict
        Dictionary with Element as keys and delta n as values.
    """
    comp_defect = structure_defect.composition
    comp_bulk = structure_bulk.composition
        
    return get_delta_atoms_from_comp(comp_defect, comp_bulk)


def get_delta_atoms_from_comp(comp_defect,comp_bulk):
    """
    Build delta_atoms dictionary starting from Composition objects.

    Parameters
    ----------
    comp_defect : Composition
        Defect structure.
    comp_bulk : Composition
        Bulk structure.

    Returns
    -------
    delta_atoms : dict
        Dictionary with Element as keys and delta n as values.
    """
    delta_atoms = {}
    for el,n in comp_defect.items():
        nsites_defect = n
        nsites_bulk = comp_bulk[el] if el in comp_bulk.keys() else 0
        delta_n = nsites_defect - nsites_bulk
        if delta_n != 0:
            delta_atoms[el] = delta_n
        
    return delta_atoms    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 18 13:12:05 2020

@author: lorenzo
"""
from itertools import product
import numpy as np
from pymatgen.core.periodic_table import Element
from pymatgen.core.composition import Composition


from .core import Chempots, chempot_ideal_gas
from .phase_diagram import PDHandler
from .reservoirs import PressureReservoirs


def get_oxygen_chempot_standard_finite_temperature(temperature,muO_reference=None):
    """
    Get value of oxygen delta mu standard (mu_0(T,Po)) at a speficic temperature.

    The data is taken from the following work: Reuter and Scheffler, “Composition, Structure, and Stability of RuO 2 (110) as a Function of Oxygen Pressure.”
    The value at a specific temperature is extracted from a linear fitting of the behaviour of mu_O with Temperature.
    
    Parameters
    ----------
    temperature : float
        Temperature in Kelvin.
    muO_ref : float
        Oxygen reference chemical potential (O2 molecule, T = 0K)

    Returns
    -------
    chempot : float
        Chemical potential at standard p0 at given temperature.

    """
    T = np.array([100,200,300,400,500,600,700,800,900,1000])
    mu_0 = np.array([-0.08, -0.17,-0.27,-0.38,-0.50,-0.61,-0.73,-0.85,-0.98,-1.10])
    
    coef = np.polyfit(T,mu_0,1)
    poly1d_fn = np.poly1d(coef)
    muO = poly1d_fn(temperature) # delta value
    if muO_reference:
        muO += muO_reference
    return muO


def get_oxygen_chempot_from_pO2(temperature=300,partial_pressure=0.2,muO_reference=None):
    """
    Get oxygen chemical potential (delta) from temperature and partial pressure.

    Parameters
    ----------
    temperature : float
        Temperature in Kelvin.
    partial_pressure : float
        Partial pressure.
    muO_ref : float
        Oxygen reference chemical potential (O2 molecule, T = 0K).
    Returns
    -------
    chempot : float
        Value of oxygen chemical potential (delta) at given T and p/p0.

    """
    muO = get_oxygen_chempot_standard_finite_temperature(temperature,muO_reference=muO_reference)
    return chempot_ideal_gas(muO,temperature,partial_pressure)  


def get_pressure_reservoirs_from_chempot_limits(composition,
                                                temperature,
                                                pressure_range=(1e-20,1e10),
                                                npoints=50,
                                                get_pressures_as_strings=False):
    """
    To be implemented
    """
    if type(composition) == str:
        composition = Composition(composition)
    partial_pressures = np.logspace(np.log10(pressure_range[0]),np.log10(pressure_range[1]),num=npoints,base=10)
    mu_standard = get_oxygen_chempot_standard_finite_temperature(temperature)
    # to be completed


def get_pressure_reservoirs_from_phase_diagram(phase_diagram,
                                               target_composition,
                                               temperature,
                                               extrinsic_chempots_range=None,
                                               pressure_range=(1e-20,1e10),
                                               interpolation_function=None,
                                               npoints=50,
                                               get_pressures_as_strings=False):
    """
    Generate Reservoirs object with a set of different chemical potentials starting from a range of oxygen partial pressure.
    The code distinguishes between 2-component and 3-component phase diagrams.
    
    In the case of a 2-comp PD the other chemical potential is calculated directly from the formation energy of the phase
    with target composition.
    In the case of a 3-comp PD the set of chemical potentials is obtained first calculating the boundary phases starting
    from the constant value of muO, then the arithmetic average between this two points in the stability diagram is taken.
    
    In the case where there are some extrinsic elements not belonging to the PD, they must be added in the 
    extrinsic_chempots_range dictionary ({element:(O_poor_chempot,O_rich_chempot)}). The chempots of extrinsic elements 
    are interpolated linearly between O_poor and O_rich values. 
    
    Parameters
    ----------
    phase_diagram : PhaseDiagram
        Pymatgen PhaseDiagram object.
    target_comp : str or Composition
        Composition of target phase.
    temperature : float
        Temperature in Kelvin.
    extrinsic_chempots_range : dict
        Dictionary with chemical potentials of elements not belonging to the PD ({element:(O_poor_chempot,O_rich_chempot)}). 
        The default is None.
    pressure_range : tuple
        Range in which to evaluate the partial pressure.
    interpolation_function : function
        Function to determine the chemical potential for the other elements (the ones that are not oxygen).

        interpolation_function(element,boundary_reservoir): The function inputs are the target 
        element and the boundary_reservoirs ({'<label>':{element:value}}). If None the mean is used.    
    npoints : int
        Number of data points to interpolate the partial pressure with. The default is 50.
    get_pressures_as_strings : bool
        Get pressure values (keys in the Reservoirs dict) as strings. The default is set to floats.

    Returns
    -------
    reservoirs
        PressureReservoirs object. The dictionary is organized as {partial_pressure:chempots}.

    """
    chempots_dict = {}
    reservoirs = {}
    pd = phase_diagram
    if type(target_composition) == str:
        target_comp = Composition(target_composition)
    else:
        target_comp = target_composition
    partial_pressures = np.logspace(np.log10(pressure_range[0]),np.log10(pressure_range[1]),num=npoints,base=10)
    mu_standard = get_oxygen_chempot_standard_finite_temperature(temperature)
    
    for idx,p in enumerate(partial_pressures):
        muO = chempot_ideal_gas(mu_standard,temperature=temperature,partial_pressure=p) #delta value
        fixed_chempot = Chempots({'O':muO})
        
        if len(pd.elements) == 2: # 2-component PD case
            pdh = PDHandler(pd)
            mu = pdh.calculate_single_chempot(target_comp, fixed_chempot)
            for el in target_comp.elements:
                if el is not Element('O'):
                    chempots = fixed_chempot.copy()
                    chempots.update({el.symbol:mu})
                
        elif len(pd.elements) == 3: # 3-component PD case
            pdh = PDHandler(pd)
            boundary_res = pdh.get_phase_boundaries_chempots(target_comp, fixed_chempot)
            for el in list(boundary_res.values())[0]:
                if interpolation_function:
                    chempots_dict[el] = interpolation_function(el,boundary_res)
                else:
                    chempots_dict[el] = np.mean(np.array([mu[el] for mu in boundary_res.values()]))
                
        elif len(pd.elements) > 3: 
            raise NotImplementedError('Not implemented for PD with more than 3 components')
        
        chempots = Chempots(chempots_dict)
        chempots_abs = chempots.get_absolute(pdh.mu_refs)
        
        if extrinsic_chempots_range:
            for el in extrinsic_chempots_range:
                mu_el_O_poor,mu_el_O_rich = extrinsic_chempots_range[el][0], extrinsic_chempots_range[el][1]
                chempots_abs[el] = mu_el_O_poor + ((mu_el_O_rich - mu_el_O_poor)/npoints)*idx 
                    
        if get_pressures_as_strings:
            p = "%.1g" % p
            p = str(p)
        else:
            p = float("{:.3e}".format(p))
        reservoirs[p] = chempots_abs

        
    return PressureReservoirs(reservoirs,temperature,phase_diagram=pd,are_chempots_delta=False)


def get_pressure_reservoirs_from_precursors(precursors,
                                            oxygen_ref,
                                            temperature,
                                            pressure_range=(1e-20,1e10),
                                            npoints=50,
                                            get_pressures_as_strings=False):
    """
    Get PressureReservoirs object starting from the oxygen reference chemical potential at 0 K and the synthesis precursors.
    Chemical potentials are found from the energies of the precursors and the oxygen chempot value 
    (uses the np.linalg.lstsq function). If the system is underdetermined the minimum-norm solution is found.

    Parameters
    ----------
    precursors : dict
        Dictionaly with formulas (str) as keys and total energies (float) as values.
    oxygen_ref : float
        Absolute chempot of oxygen at 0K.
    temperature : float
        Temperature in K.
    pressure_range : tuple
        Range in which to evaluate the partial pressure . The default is from 1e-20 to 1e10.
    npoints : int
        Number of data points to interpolate the partial pressure with. The default is 50.
    get_pressures_as_strings : bool
        Get pressure values (keys in the Reservoirs dict) as strings. The default is set to floats.

    Returns
    -------
    pressure_reservoirs : PressureReservoirs
        PressureReservoirs object.

    """
    row_elements = []
    for formula in precursors.keys():
        comp = Composition(formula)
        for element in comp.elements:
            el = element.symbol
            if el not in row_elements and el != 'O':
                row_elements.append(el)
    A = []
    N = len(row_elements)
    for formula in precursors.keys():
        a = np.zeros(N)
        comp = Composition(formula)
        for element,coeff in comp.items():
            el = element.symbol
            if el != 'O':
                a[row_elements.index(el)] = coeff
        A.append(a)

    reservoirs = get_oxygen_pressure_reservoirs(oxygen_ref=oxygen_ref,
                                            temperature=temperature,
                                            pressure_range=pressure_range,
                                            npoints=npoints,
                                            get_pressures_as_strings=get_pressures_as_strings)
    for chempots in reservoirs.values():
        muO = chempots['O']
        B = []
        for formula,energy in precursors.items():
            comp = Composition(formula)
            b = energy - comp['O']*muO
            B.append(b)
        X = np.linalg.lstsq(A, B, rcond=None)[0]
        for index,el in enumerate(row_elements):
            chempots[el] = X[index]
    
    return reservoirs


def get_oxygen_pressure_reservoirs(oxygen_ref,temperature,pressure_range=(1e-20,1e10),npoints=50,get_pressures_as_strings=False):
    """
    Get PressureReservoirs object for oxygen starting from the reference value.

    Parameters
    ----------
    oxygen_ref : float
        Absolute chempot of oxygen at 0K.
    temperature : float
        Temperature.
    pressure_range : tuple
        Range in which to evaluate the partial pressure . The default is from 1e-20 to 1e10.
    npoints : int
        Number of data points to interpolate the partial pressure with. The default is 50.
    get_pressures_as_strings : bool
        Get pressure values (keys in the Reservoirs dict) as strings. The default is set to floats.

    Returns
    -------
    pressure_reservoirs : PressureReservoirs
        PressureReservoirs object.

    """
    reservoirs = {}
    mu_refs = Chempots({'O':oxygen_ref})
    partial_pressures = np.logspace(np.log10(pressure_range[0]),np.log10(pressure_range[1]),num=npoints,base=10)
    mu_standard = get_oxygen_chempot_standard_finite_temperature(temperature)
    for p in partial_pressures:
        mu = {}
        muO = chempot_ideal_gas(mu_standard,temperature=temperature,partial_pressure=p)
        mu.update({'O':oxygen_ref + muO})
        if get_pressures_as_strings:
            p = "%.1g" % p
            p = str(p)
        else:
            p = float("{:.3e}".format(p))
        reservoirs[p] = Chempots(mu)
    
    return PressureReservoirs(reservoirs,temperature,phase_diagram=None,
                              mu_refs=mu_refs,are_chempots_delta=False)

    

        
def get_barycenter_chemical_potentials_absolute(composition,
                                                energy,
                                                oxygen_chempot_absolute,
                                                mu_refs,
                                                min_absolute_chempots=None,
                                                max_absolute_chempots=None):
    """

    Compute the barycenter of the feasible region for relative chemical potentials, constrained by:
    
    - fixed absolute chemical potential of oxygen
    - Total energy of target phase
    - lower and/or upper chemical potential limits for each element.
    
    Parameters
    ----------
    composition: str or pymatgen.core.Composition
        Target composition.
    formation_energy: float
        Formation energy (eV/f.u.)
    oxygen_mu_relative: float
        fixed chemical potential of oxygen relative to the oxygen molecule.
    min_relative_chempots: dict or Chempots
        Lower limit of chemical potentials, relative values ({element:chempot}) 
    max_relative_chempots: dict or Chempots
        Higher limit of chemical potentials, relative values ({element:chempot})
    
    Returns:
        Dictionary with chemical potentials, taken from the center of the
        allowed N-1 dimensional hyperplane.

    """ 
    if isinstance(composition, str):
        composition = Composition(composition)

    formation_energy = energy - sum([number*mu_refs[el.symbol] for el,number in composition.items()])
    oxygen_chempot_relative = oxygen_chempot_absolute - mu_refs['O']
    min_relative_chempots = Chempots(min_absolute_chempots).get_referenced(mu_refs) if min_absolute_chempots else None
    max_absolute_chempots = Chempots(max_absolute_chempots).get_referenced(mu_refs) if max_absolute_chempots else None
    
    barycenter_chempots_relative = get_barycenter_chemical_potentials_relative(composition=composition,
                                                   formation_energy=formation_energy,
                                                   oxygen_chempot_relative=oxygen_chempot_relative,
                                                   min_relative_chempots=min_relative_chempots,
                                                   max_relative_chempots=max_absolute_chempots)
    
    return Chempots(barycenter_chempots_relative).get_absolute(mu_refs)
        

    

def get_barycenter_chemical_potentials_relative(composition,
                                                formation_energy,
                                                oxygen_chempot_relative,
                                                min_relative_chempots=None,
                                                max_relative_chempots=None):
    """
    
    Compute the barycenter of the feasible region for relative chemical potentials, constrained by:

    - fixed relative chemical potential of oxygen
    - Formation energy of target phase
    - lower and/or upper chemical potential limits for each element
    
    Parameters
    ----------
    composition: str or pymatgen.core.Composition
        Target composition.
    formation_energy: float
        Formation energy (eV/f.u.)
    oxygen_mu_relative: float
        fixed chemical potential of oxygen relative to the oxygen molecule.
    min_relative_chempots: dict or Chempots
        Lower limit of chemical potentials, relative values ({element:chempot}) 
    max_relative_chempots: dict or Chempots
        Higher limit of chemical potentials, relative values ({element:chempot})
    
    Returns
    -------
        Dictionary with chemical potentials, taken from the center of the
        allowed N-1 dimensional hyperplane.
        
    """
    muO_relative = oxygen_chempot_relative
    if isinstance(composition, str):
        composition = Composition(composition)
    
    el_amount_dict = composition.get_el_amt_dict()
    if "O" not in el_amount_dict:
        raise ValueError("Oxygen must be present in the composition.")

    n_O = el_amount_dict['O']
    non_oxygen_elements = [el for el in el_amount_dict if el !='O']
    constraint = formation_energy - n_O * muO_relative
    chempot_boundaries = {}
    for el in non_oxygen_elements:
        if min_relative_chempots and el in min_relative_chempots.keys():
            mu_min = min_relative_chempots[el]
        else:
            mu_min = float("-inf")
        if max_relative_chempots and el in max_relative_chempots.keys():
            mu_max = max_relative_chempots[el]
        else:
            mu_max = 0.0
        chempot_boundaries[el] = (mu_min,mu_max)
    
    allowed_vertices = []
    # iterate over all possible permutations of chempot ranges
    for node in product(*[chempot_boundaries[el] for el in non_oxygen_elements]):
        for idx in range(len(non_oxygen_elements)):
            mu_values = list(node)
            el_free = non_oxygen_elements[idx]
            n_el_free = el_amount_dict[el_free]
            
            fixed_sum = sum(el_amount_dict[non_oxygen_elements[i]] * mu_values[i]
                            for i in range(len(non_oxygen_elements)) if i != idx)
            
            mu_free = (constraint - fixed_sum) / n_el_free  # solve free chemical potentials
            
            mu_min, mu_max = chempot_boundaries[el_free]
            if mu_min <= mu_free <= mu_max:  #if solved chempot is within limits, we add vertex to N-1 chempot hyperplane
                mu_values[idx] = mu_free
                vertex = dict(zip(non_oxygen_elements, mu_values))
                allowed_vertices.append(vertex)
                
    if not allowed_vertices:
        raise ValueError("No feasible chemical potential points found under the given constraints.")
        
    barycenter = {el:0 for el in non_oxygen_elements}
    for vertex in allowed_vertices:
        for el, mu in vertex.items():
            barycenter[el] += mu
    for el in barycenter.keys():
        barycenter[el] /= len(allowed_vertices)
    barycenter['O'] = muO_relative
    
    return barycenter        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 22 13:56:05 2023

@author: villa
"""
from scipy import constants
from scipy.constants import physical_constants 
import numpy as np
import pandas as pd

from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.core.units import kb

from defermi.analysis import DefectsAnalysis
from defermi.chempots.core import Chempots
from defermi.defects import Vacancy
from defermi.entries import DefectEntry
from defermi.thermodynamics import DefectThermodynamics
from defermi.tools.utils import get_object_from_json

from defermi.testing.core import DefermiTest


e = constants.e

        
class TestDefectThermodynamics(DefermiTest):
    
    @classmethod
    def setUpClass(cls):
        self = cls()
        bulk_volume = 909.0035
        d =[{'name': 'Int_O',
            'charge': -2.0,
            'multiplicity': 1,
            'energy_diff': 4.543,
            'bulk_volume': bulk_volume},
            {'name': 'Int_O',
            'charge': -1.0,
            'multiplicity': 1,
            'energy_diff': 1.7746,
            'bulk_volume': bulk_volume},
            {'name': 'Int_O',
            'charge': 0.0,
            'multiplicity': 1,
            'energy_diff': -0.3334,
            'bulk_volume': bulk_volume},
            {'name': 'Sub_P_on_Si',
            'charge': 0.0,
            'multiplicity': 1,
            'energy_diff': 5.6674,
            'bulk_volume': bulk_volume},
            {'name': 'Sub_P_on_Si',
            'charge': 1.0,
            'multiplicity': 1,
            'energy_diff': -1.5959,
            'bulk_volume': bulk_volume},
            {'name': 'Vac_O',
            'charge': 0.0,
            'multiplicity': 1,
            'energy_diff': 9.8442,
            'bulk_volume': bulk_volume},
            {'name': 'Vac_O',
            'charge': 1.0,
            'multiplicity': 1,
            'energy_diff': 8.3845,
            'bulk_volume': bulk_volume},
            {'name': 'Vac_O',
            'charge': 2.0,
            'multiplicity': 1,
            'energy_diff': 4.0104,
            'bulk_volume': bulk_volume},
            {'name': 'Vac_Si',
            'charge': -5.0,
            'multiplicity': 1,
            'energy_diff': 38.4723,
            'bulk_volume': bulk_volume},
            {'name': 'Vac_Si',
            'charge': -4.0,
            'multiplicity': 1,
            'energy_diff': 31.4767,
            'bulk_volume': bulk_volume},
            {'name': 'Vac_Si',
            'charge': -3.0,
            'multiplicity': 1,
            'energy_diff': 28.4873,
            'bulk_volume': bulk_volume}]
        
        df = pd.DataFrame(d)
        vbm = 1.0394
        band_gap = 5.968500000000001
        da = DefectsAnalysis.from_dataframe(df,vbm=vbm,band_gap=band_gap)
        cls.da = da
        mu_O = -6
        mu_P = -5.4133
        mu_Si = -5.4224
        cls.chempots = Chempots({'Si':mu_Si,'P':mu_P,'O':mu_O},ndecimals=2)
        cls.dos = get_object_from_json(CompleteDos, self.get_testfile_path('SiO2-defects/Bulk-2x2x2-supercell/complete_dos.json'))

    
    def test_brouwer_diagram_from_precursors(self):
        da = self.da.filter_entries(elements=['P'],exclude=True)
        da.plot_brouwer_diagram(
                        bulk_dos=self.dos,
                        temperature=1000,
                        precursors={'SiO2':-23.69},
                        oxygen_ref=-4.95,
                        pressure_range=(1e-20,1e10),
                        npoints=50,
                        xtol=1e-15)
        
        data = da.thermodata
        conc = data.defect_concentrations[5]

        actual = data.partial_pressures[37]
        desired = 449.8
        self.assert_all_close(actual,desired, rtol=1e-03)

        actual = conc.select_concentrations(name='Vac_O',charge=0)[0]['conc']
        desired = 20033806666.685966
        self.assert_all_close(actual,desired, rtol=1e-03)
        
        conc = data.defect_concentrations[21]
        actual = conc.select_concentrations(name='Int_O',charge=-2)[0]['conc']
        desired = 152050.744715061
        self.assert_all_close(actual, desired, rtol=1e-03)

        actual = data.fermi_levels[12]
        desired = 3.0835355572700505
        self.assert_all_close(actual, desired, rtol=1e-03)

    
    def test_doping_diagram(self):
        da = self.da
        da.plot_doping_diagram(
                        variable_defect_specie='P',
                        concentration_range=(1,1e20),
                        chemical_potentials=self.chempots,
                        bulk_dos=self.dos,
                        temperature=1000,
                        npoints=50,
                        xtol=1e-100)
        
        data = da.thermodata
        self.assertEqual(data.variable_defect_specie, 'P')

        actual = data.variable_concentrations[7]
        desired = 719.6856730011522
        self.assert_all_close(actual,desired, rtol=1e-03)

        actual = data.defect_concentrations[7].elemental['P']
        self.assert_all_close(actual, desired, rtol=1e-03)

        conc = data.defect_concentrations[22]
        actual = conc.select_concentrations(name='Int_O',charge=-1)[0]['conc']
        desired = 175.7802582858784
        self.assert_all_close(actual, desired, rtol=1e-03)

        actual = data.fermi_levels[13]
        desired = 2.7925563006401064
        self.assert_all_close(actual, desired, rtol=1e-03)

    
    def test_custom_functions(self):
        e1 = DefectEntry(Vacancy('O',charge=2,bulk_volume=800,multiplicity=1),energy_diff=7)
        e2 = DefectEntry(Vacancy('Sr',charge=-2,bulk_volume=800,multiplicity=1),energy_diff=8)
        da = DefectsAnalysis([e1,e2],vbm=0,band_gap=2)
        chempots = {'O':-4.95,'Sr':-2}
        mdos = {'m_eff_e':0.5,'m_eff_h':0.4}

        def custom_eform(entry,vbm=None,chemical_potentials=None,fermi_level=0,temperature=300,**kwargs):
                formation_energy = entry.energy_diff + entry.charge*(vbm+fermi_level) 
                if chemical_potentials:
                    chempot_correction = -1 * sum([entry.delta_atoms[el]*chemical_potentials[el] for el in entry.delta_atoms])
                else:
                    chempot_correction = 0
                    
                formation_energy = formation_energy + chempot_correction
                test_quantity = kwargs['test'] if kwargs and 'test' in kwargs.keys() else 0
                return formation_energy - 1/500 *temperature + test_quantity       
        

        def custom_dconc(entry,vbm=0,chemical_potentials=None,temperature=0,fermi_level=0,per_unit_volume=True,eform_kwargs={},**kwargs):
                from defermi.entries import fermi_dirac
                n = entry.defect.site_concentration_in_cm3 if per_unit_volume else entry.multiplicity 
                if per_unit_volume:
                    n = n * (1+1/500*temperature)
                eform = entry.formation_energy(
                                        vbm=vbm,
                                        chemical_potentials=chemical_potentials,
                                        fermi_level=fermi_level,
                                        temperature=temperature,
                                        **eform_kwargs)
                
                test_quantity = kwargs['test_conc'] if kwargs and 'test_conc' in kwargs.keys() else 1
                return  n * fermi_dirac(eform,temperature) * test_quantity
        
        da.set_formation_energy_functions(function=custom_eform,name='Vac_O')
        da.set_defect_concentration_functions(function=custom_dconc,name='Vac_Sr')

        da.plot_brouwer_diagram(mdos,1000,precursors={'SrO':-10},oxygen_ref=-4.95,xtol=1e-100)
        data = da.thermodata

        conc = data.defect_concentrations[35]
        actual = conc.select_concentrations(name='Vac_O')[0]['conc']
        desired = 537497932157133.1
        self.assert_all_close(actual, desired, rtol=1e-05)

        actual = data.carrier_concentrations[28][0]
        desired = 28853383959330.863
        self.assert_all_close(actual, desired, rtol=1e-05)


### more complex functions with quenched species and external defects to be implemented


        
        
        



import numpy as np
from monty.json import MSONable
from pymatgen.core.periodic_table import Element
import copy


class Chempots(MSONable):
    
    def __init__(self,chempots_dict,ndecimals=None):
        """
        Class to handle set of chemical potentials. Behaves like a python dictionary.
        The dictionary needs to be set with element symbols as keys and chemical potentials 
        as values.

        Parameters
        ----------
        chempots_dict : dict
            Dictionary of chemical potentials in the format {el:value}.
        ndecimals : int
            Round the chemical potentials to this number of decimals. If None
            the numbers are left untouched.
            
        """
        if ndecimals:
            self._mu = {el:round(v,ndecimals) for el,v in chempots_dict.items()}
        else:
            self._mu = chempots_dict
    
    @property
    def mu(self):
        """
        Dictionary with chemical potentials
        """
        return self._mu

    def __str__(self):
        return self.mu.__str__()
    
    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.mu)

    def __iter__(self):
        return self.mu.keys().__iter__()

    def __getitem__(self,el):
        return self.mu[el]

    def __setitem__(self,el,value):
        self.mu[el] = value
        return
    
    def __delitem__(self,el):
        del self.mu[el]
        return
    
    def __eq__(self, other):
        if isinstance(other, dict):
            return self.mu == other
        elif isinstance(other, Chempots):
            return self.mu == other.mu
        else:
            return False

    def keys(self):
        return self.mu.keys()

    def values(self):
        return self.mu.values()
    
    def items(self):
        return self.mu.items()

    def copy(self):
        return Chempots(copy.deepcopy(self.mu))
    
    def update(self, other):
        if isinstance(other, dict):
            for key, value in other.items():
                self.mu[key] = value
        elif other.__class__.__name__ == 'Chempots':
            for key, value in other.items():
                self.mu[key] = value
        else:
            raise ValueError('Chempots object can be only updated with dict or Chempots objects')
    
    
    def as_dict(self):
        """
        Json-serializable dict representation of a Chempots object. 
        
        Returns
        -------
        d : dict
            Json-serializable dict of a Chempots object.

        """
        d = {
        "@module": self.__class__.__module__,
        "@class": self.__class__.__name__,
        "chempots":self.mu
        }
        return d
    
    @staticmethod
    def from_dict(d):
        """
        Constructor of Chempots object from dictionary representation.
        
        Parameters
        ----------
        d : dict
        
        Returns
        -------
        chempots : Chempots
            Chempots object.

        """
        chempots = d["chempots"]
        return Chempots(chempots)
    
    @staticmethod
    def from_pmg_elements(d):
        """
        Construct a Chempots object from a dictionary in the format {Element:value}. Useful to
        interface with pymatgen which uses Element class to address elements instead of just their symbols.

        Parameters
        ----------
        d : dict
            Dictionary in the format {Element:value}.

        Returns
        -------
        chempots : Chempots
            Chempots object.

        """
        return Chempots({el.symbol:value for el,value in d.items()})
    
    def to_pmg_elements(self):
        """
        Convert the Chempots object to a dictionary in the format {Element:value}. 
        Useful to interface with pymatgen which uses Element class to address elements 
        instead of just their symbols.

        Returns
        -------
        d : dict
            Dictionary in the format {Element:value}

        """
        return {Element(el):value for el,value in self.mu.items()}
    
    
    def get_absolute(self,mu_refs):
        """
        Get Chempots object with chemical potentials converted to absolute values (mu + mu_ref).

        Parameters
        ----------
        mu_refs : dict or Chempots
            Chempots object or just dictionary with elemental chemical potentials (mu0).

        Returns
        -------
        chempots : Chempots
            Chempots values converted to absolute (mu + mu_ref).

        """
        return Chempots({el:self.mu[el] + mu_refs[el] for el in self.mu})
        
    
    def get_referenced(self,mu_refs):
        """
        Get Chempots object with chemical potentials converted to referenced values (mu - mu_ref).

        Parameters
        ----------
        mu_refs : dict or Chempots
            Chempots object or just dictionary with elemental chemical potentials (mu0).

        Returns
        ------
        chempots : Chempots
            Chempots values converted to referenced (mu - mu_ref).

        """
        return Chempots({el:self.mu[el] - mu_refs[el] for el in self.mu})


def chempot_ideal_gas(mu0, temperature,partial_pressure,coeff=2):
    """
    Get chemical potential at a given temperature and partial pressure. 
    The chemical potential in standard conditions (mu0) has to be know.

    Parameters
    ----------
    mu0 : float
        Value of chemical potential in standard conditions.
    temperature : float
        Temperature in Kelvin.
    partial_pressure : float
        Partial pressure over standard pressure (p/p0).
    coeff : int
        Stochiometric coefficient. Default is 2 for oxygen.

    Returns
    -------
    chempot : float
        Value of chemical potential at given T and p/p0.

    """
    kb = 8.6173324e-5  # eV / K
    chempot = mu0 + (1/coeff)*kb*temperature*np.log(partial_pressure)
    return chempot



            
    

    




import warnings
import os.path as op
import importlib
import matplotlib.pyplot as plt
import math


from pymatgen.io.vasp.outputs import Outcar, Vasprun

from pymatgen.analysis.defects.corrections.kumagai import _check_import_pydefect 
from pymatgen.analysis.defects.utils import CorrectionResult, get_zfile

from ..tools.structure import is_site_in_structure_coords
from ..structure import defect_finder

from pathlib import Path



    
def get_kumagai_correction(
                        defect_path,
                        bulk_path,
                        charge,
                        dielectric_tensor,
                        initial_structure=False,
                        get_correction_data=True,
                        get_plot=True,
                        **kwargs):
    """
    Compute Kumagai corrections (extended FNV scheme) from VASP calculation paths.

    Parameters
    ----------
    defect_path : str
        Path of defect calculation with vasprun.xml and OUTCAR files.
    bulk_path : str
        Path of bulk calculation with vasprun.xml and OUTCAR files.
    charge : int
        Charge of defect calculation
    dielectric_tensor : int,float 3x1 array or 3x3 array
        Dielectric tensor (or constant). Types accepted are int,float 3x1 array or 3x3 array.
    initial_structure : bool
        Use initial structure of defect calculation for correction computation.
        Useful to compute correction on unrelaxed structure.
    get_correction_data : bool
        Return pymatgen's CorrectionResult object. If False only the correction value is returned.
    get_plot : bool
        Get Matplotlib object with plot. The default is False. 
    kwargs : dict
        Kwargs to pass to pydefect `make_efnv_correction`
    
    Returns
    -------
    correction, ax : tuple
        CorrectionResult object if get_correction_data is set to True, else just the float with correction value, matplotlib axis object  
        
    """
    defect_structure = _get_structure_with_pot_pmg(defect_path,initial_structure=initial_structure)
    bulk_structure = _get_structure_with_pot_pmg(bulk_path,initial_structure=False)
    correction = get_kumagai_correction_from_structures(
                                            defect_structure_with_potentials=defect_structure,
                                            bulk_structure_with_potentials=bulk_structure,
                                            charge=charge,
                                            dielectric_tensor=dielectric_tensor,
                                            **kwargs)
    
    corr = correction if get_correction_data else correction.correction_energy
    if get_plot:
        from pydefect.corrections.site_potential_plotter import SitePotentialMplPlotter
        SitePotentialMplPlotter.from_efnv_corr(
                                            title=defect_structure.composition,
                                            efnv_correction=correction.metadata['efnv_corr']
                                            ).construct_plot()
        ax = plt.gca()
        return corr, ax
    else:
        return corr 



def get_kumagai_correction_from_structures(
                                        defect_structure_with_potentials,
                                        bulk_structure_with_potentials,
                                        charge,
                                        dielectric_tensor,
                                        **kwargs):
    """
    Get Kumagai (extended FNV) correction from Structure objects with 
    site potentials stored into site_properties. Not recommended to use
    directly, refer to `get_kumagai_correction`.
    """
    dielectric_tensor = _convert_dielectric_tensor(dielectric_tensor)
    correction =  _get_efnv_correction_pmg_fixed(
                            charge=charge,
                            defect_structure=defect_structure_with_potentials,
                            bulk_structure=bulk_structure_with_potentials,
                            dielectric_tensor=dielectric_tensor,
                            **kwargs)
    return correction




def _get_efnv_correction_pmg_fixed(
    charge,
    defect_structure,
    bulk_structure,
    dielectric_tensor,
    **kwargs):
    """
    Fixed code from `pymatgen.analysis.defects.corrections.kumagai.get_efnv_correction`,
    defect_potentials and site_potentials variables were inverted.
    
    Returns the Kumagai/Oba EFNV correction for a given defect.

    Parameters
    ----------
    charge : int or float
        Charge of the defect.
    defect_structure : Structure
        Defect structure.
    bulk_structure : Structure
        Bulk structure.
    dielectric_tensor : np.array
        Dielectric tensor.
    kwargs : dict
        Kwargs to pass to `make_efnv_correction`.
    """
    from pydefect.analyzer.calc_results import CalcResults
    from pydefect.cli.vasp.make_efnv_correction import make_efnv_correction

    # ensure that the structures have the "potential" site property
    defect_potentials = [site.properties["potential"] for site in defect_structure]
    bulk_potentials = [site.properties["potential"] for site in bulk_structure]

    defect_calc_results = CalcResults(
        structure=defect_structure,
        energy=math.inf,
        magnetization=math.inf,
        potentials=defect_potentials,
    )
    bulk_calc_results = CalcResults(
        structure=bulk_structure,
        energy=math.inf,
        magnetization=math.inf,
        potentials=bulk_potentials,
    )

    efnv_corr = make_efnv_correction(
        charge=charge,
        calc_results=defect_calc_results,
        perfect_calc_results=bulk_calc_results,
        dielectric_tensor=dielectric_tensor,
        **kwargs,
    )

    return CorrectionResult(
        correction_energy=efnv_corr.correction_energy,
        metadata={"efnv_corr": efnv_corr},
    )


def _get_structure_with_pot_pmg(directory,initial_structure=False):
    """
    Modified function from pymatgen.analysis.defects-corrections.kumagai 
    to allow for initial structure import

    Reads vasprun.xml and OUTCAR files in a directory.

    Parameters
    ----------
    directory : str
        Directory containing vasprun.xml and OUTCAR files.

    Returns
    -------
    structure : Structure
        Structure with "potential" site property.
    """
    _check_import_pydefect()
    from pydefect.analyzer.calc_results import CalcResults

    d_ = Path(directory)
    f_vasprun = get_zfile(d_, "vasprun.xml")
    f_outcar = get_zfile(d_, "OUTCAR")
    vasprun = Vasprun(f_vasprun,parse_dos=False,parse_potcar_file=False,parse_eigen=False)
    outcar = Outcar(f_outcar)

    structure = vasprun.structures[0] if initial_structure else vasprun.final_structure

    calc = CalcResults(
        structure=structure,
        energy=outcar.final_energy,
        magnetization=outcar.total_mag or 0.0,
        potentials=[-p for p in outcar.electrostatic_potential],
        electronic_conv=vasprun.converged_electronic,
        ionic_conv=vasprun.converged_ionic,
    )

    return calc.structure.copy(site_properties={"potential": calc.potentials})


def _convert_dielectric_tensor(dielectric):
    import numpy as np
    if not type(dielectric) in [float,int]:
        dielectric = np.array(dielectric)
        if dielectric.shape == (3,):
            dielectric = np.diag(dielectric)
        elif dielectric.shape != (3, 3):
            raise ValueError("Dielectric tensor can be int/float, a 3x1 array with diagonal components of dielectric tensor, or 3x3 matrix")
    else:
        dielectric = np.eye(3) * dielectric

    return dielectric





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 14:35:23 2021

@author: villa
"""

import numpy as np
import pandas as pd
import matplotlib



def get_elastic_dipole_tensor(stress_defect,stress_bulk,bulk_volume,corrections={}):
    """
    Calculate elastic dipole tensor from stresses.

    Parameters
    ----------
    stress_defect : np.array
        Stresses from defect calculation in kbar (VASP output)
    stress_bulk : np.array
        Stresses from bulk calculation in kbar (VASP output)
    bulk_volume : float
        Cell volume of bulk calculation in A°^3.
    corrections : bool
        Add correction terms to the residual stress tensor.

    Returns
    -------
    dipole_tensor : np.array
        Elastic dipole tensor in eV.
    """    
    res_stress = np.array(stress_defect) - np.array(stress_bulk)
    dipole_tensor = -1*bulk_volume*res_stress #sign is inverted with respect to VASP output
    dipole_tensor += sum([v for k,v in corrections.items()]) 
    return dipole_tensor


def get_relaxation_volume(stress_defect,stress_bulk,bulk_modulus,bulk_volume,corrections={}):
    """
    Calculate relaxation volume from stresses.

    Parameters
    ----------
    stress_defect : np.array
        Stresses from defect calculation in kbar (VASP output)
    stress_bulk : np.array
        Stresses from bulk calculation in kbar (VASP output)
    bulk_volume : float
        Cell volume of bulk calculation in A°^3.
    bulk_modulus : float
        Bulk modulus in GPa.
    corrections : bool
        Add correction terms to the residual stress tensor.
        
    Returns
    -------
    rel_volume : float
        Relaxation volume in A°^3.
    """
    bulk_modulus = bulk_modulus*10 # from GPa to kbar
    dipole_tensor = get_elastic_dipole_tensor(
                                            stress_defect=stress_defect,
                                            stress_bulk=stress_bulk,
                                            bulk_volume=bulk_volume,
                                            corrections=corrections)
    
    pressure = np.trace(dipole_tensor)/3
    rel_volume = pressure/bulk_modulus
    return rel_volume



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 11 14:33:12 2020

@author: villa
"""

from abc import ABCMeta
from monty.json import MSONable, MontyDecoder, MontyEncoder, jsanitize
import numpy as np
import warnings
import os

from pymatgen.core.units import kb

from .structure import defect_finder
from .elasticity import get_relaxation_volume
from .tools.utils import get_charge_from_computed_entry



class DefectEntry(MSONable,metaclass=ABCMeta):
    
    def __init__(self,
                defect,
                energy_diff=None,
                corrections={},
                data=None,
                formation_energy_function=None,
                defect_concentration_function=None):
        """
        Object to store the results of a defect calculation. 
        This object is usually created automatically when importing results, either from directories or
        DataFrame.
        
        Parameters
        ----------
        defect : Defect
            Defect object (Vacancy, Interstitial, Substitution, Polaron or DefectComplex)
        energy_diff : float
            Difference btw energy of defect structure and energy of pristine (bulk) structure in eV.
        corrections : dict
            Dictionary of corrections in eV to apply to defect formation energy.
            All values will be added to the defect formation energy. Keys indicate the correction type.     
        data : dict
            Store additional data in dict format.
        label : str
            Additional label to add to defect specie.
        formation_energy_function : function
            Custom function for calculation of defect formation energy.
            Check documentation in `formation_energy` function for more details.
        defect_concentration_function : function
            Custom function for calculation of defect concentration.
            Check documentation in `defect_concentration` function for more details.
        """
        self._defect = defect
        self._energy_diff = energy_diff
        self._corrections = corrections
        self._data = data if data else {}
        self._formation_energy_function = formation_energy_function
        self._defect_concentration_function = defect_concentration_function
    
    def __repr__(self):
        return "DefectEntry: Name=%s, Charge=%i" %(self.name,self.charge)

    def __str__(self):
        output = [
            "DefectEntry",
            "Defect: %s" %(self.defect.__str__()),
            "Energy: %.4f" %self.energy_diff,
            "Corrections: %.4f" %sum([v for v in self.corrections.values()]),
            "Charge: %.1f" %self.charge,
            "Multiplicity: %i" %self.multiplicity,
            "Data: %s" %list(self.data.keys()),
            "Name: %s" %self.name,
            "\n"
            ]
        return "\n".join(output)

    @property
    def bulk_structure(self):
        """
        Structure of pristine material.
        """
        return self.defect.bulk_structure   

    @property
    def charge(self):
        if self.defect.charge is None:
            warnings.warn('Charge is not set in Defect object, setting charge to 0')
            self.defect.set_charge(0)
        return self.defect.charge
    
    @property
    def corrections(self):
        """
        Dictionary with corrections to the defect energy. Its values will be added to the formation energy.
        """
        return self._corrections
    
    @property
    def data(self):
        """
        Dictionary containing additional data.
        """
        return self._data

    @property
    def defect(self):
        """
        Defect object.
        """
        return self._defect  
    
    @property
    def defect_specie(self):
        """
        Species involved in defect.
        """
        return self.defect.specie
    
    @property
    def defect_type(self):
        """
        Defect type.
        """
        return self.defect.type
    
    @property
    def delta_atoms(self):
        """
        Dictionary with Element as keys and particle difference between defect structure
        and bulk structure as values.
        """
        return self.defect.delta_atoms
    
    @property
    def energy_diff(self):
        return self._energy_diff
    
    @property
    def label(self):
        return self.defect.label

    @property
    def multiplicity(self):
        if self.defect.multiplicity is None:
            warnings.warn('Multiplicity is not set in Defect object, setting multiplicity to 1.')
            self.defect.set_multiplicity(1)
        return self.defect.multiplicity
                
    @property
    def name(self):
        return self.defect.name
        
    @property
    def structure(self):
        return self.defect.defect_structure
    
    @property
    def symbol(self):
        return self.defect.symbol

    @property
    def symbol_charge(self):
        """
        Defect symbol with charge formatted with numbers.
        """
        return self.defect.symbol_with_charge

    @property
    def symbol_kroger(self):
        """
        Defect symbol with charge formatted with Kröger-Vink notation.
        """
        return self.defect.symbol_with_charge_kv
    

    @property
    def formation_energy_function(self):
        return self._formation_energy_function
    
    @property
    def defect_concentration_function(self):
        return self._defect_concentration_function
    
    def set_charge(self,new_charge=0):
        """
        Set defect charge
        """
        self.defect.set_charge(new_charge)
        return 
    
    def set_corrections(self,**kwargs):
        """
        Set defect corrections. Pass either dict or kwargs.
        """
        if self._corrections is None:
            self._corrections = {}
        for k,v in kwargs.items():
            self._corrections[k] = v
        return

    def set_data(self,new_data):
        """
        Set data dictionary
        """
        self._data = new_data
    
    def set_multiplicity(self,new_multiplicity=None):
        """
        Sets the Multiplicity of the defect.
        If `new_multiplicity` is not provided it is determined 
        with `self.defect.get_multiplicity`
        """
        self.defect.set_multiplicity(new_multiplicity)
    
    def set_label(self,new_label):
        """
        Set defect label.
        """
        self.defect.set_label(new_label)


    def set_defect_concentration_function(self, new_function):
        """
        Set custom defect concentration function.
        """
        self._defect_concentration_function = new_function
    
    def set_formation_energy_function(self, new_function):
        """
        Set custom defect concentrations function.
        """
        self._formation_energy_function = new_function

    def reset_defect_concentration_function(self):
        """
        Reset formation energy function to default.
        """
        self._defect_concentration_function = None
    
    def reset_formation_energy_function(self):
        """
        Reset formation energy function to default.
        """
        self._formation_energy_function = None

    
    def as_dict(self):
        """
        Returns:
            Json-serializable dict representation of DefectEntry.
        """
        d = {}
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        d['defect'] = self.defect.as_dict()
        d['energy_diff'] = self.energy_diff
        d['corrections'] = jsanitize(self.corrections)
        d['data'] = jsanitize(self.data)
        return d


    @classmethod
    def from_dict(cls,d):
        """
        Reconstitute a DefectEntry object from a dict representation created using
        as_dict().

        Parameters
        ----------
        d : dict
            Dictionary representation of DefectEntry.

        Returns
        -------
        entry : DefectEntry
            DefectEntry object

        """
        defect = MontyDecoder().process_decoded(d['defect'])
        energy_diff = d['energy_diff']
        corrections = d['corrections']
        data = d['data']
        return cls(defect=defect,energy_diff=energy_diff,corrections=corrections,data=data)


    @staticmethod
    def from_computed_entries(
                            computed_entry_defect,
                            computed_entry_bulk,
                            corrections,
                            multiplicity=1,
                            data=None,
                            label=None,
                            initial_structure=False,
                            **kwargs):
        """
        Generate DefectEntry object from Pymatgen's ComputedStructureEntry objects.

        Parameters
        ----------
        computed_entry_defect : VaspJob
            ComputedStructureEntry of the defect calculation.
        computed_entry_bulk : VaspJob
            ComputedStructureEntry of the bulk calculation.
        corrections : dict
            Dict of corrections for defect formation energy. All values will be summed and
            added to the defect formation energy.
        multiplicity : int
            Multiplicity of defect within the supercell. 
            If set to None is attempted to be determined automatically.
        data : dict
            Store additional data in dict format.
        label : str
            Additional label to add to defect specie. Does not influence non equilibrium calculations.
        initial_structure : bool
            Use initial structure for defect recognition. Useful when relaxations are large and 
            defect_finder struggles to find the right defects.
        kwargs : dict
            Kwargs to pass to `defect_finder`.

        Returns
        -------
        entry : DefectEntry
            DefectEntry object

        """ 
        entry_df, entry_bulk = computed_entry_defect,computed_entry_bulk
        charge = get_charge_from_computed_entry(entry_df)
        energy_diff = entry_df.energy - entry_bulk.energy
        if initial_structure:
            defect_structure = entry_df.data['ionic_steps'][0]['structure']
        else:
            defect_structure = entry_df.structure
        
        return DefectEntry.from_structures(defect_structure=defect_structure,
                                           bulk_structure=entry_bulk.structure,
                                           energy_diff=energy_diff,corrections=corrections,
                                           charge=charge,multiplicity=multiplicity,data=data,
                                           label=label,**kwargs)
    

    @staticmethod
    def from_structures(
                    defect_structure,
                    bulk_structure,
                    energy_diff,
                    corrections,
                    charge=0,
                    multiplicity=1,
                    data=None,
                    label=None,
                    **kwargs):
        """
        Generate DefectEntry object from Structure objects.

        Parameters
        ----------
        defect_structure : Structure
            Defect structure.
        bulk_structure : Structure
            Bulk structure.
        energy_diff : float 
            Difference btw energy of defect structure and energy of pristine structure
        corrections : dict
            Dict of corrections in eV for defect formation energy. All values will be summed and
            added to the defect formation energy.  
        charge : int or float
            Charge of the defect system. The default is 0.
        multiplicity : int
            multiplicity of defect within the supercell. 
            If set to None is attempted to be determined automatically with Pymatgen. The default is 1.
        data : dict
            Store additional data in dict format.
        label : str
            Additional label to add to defect specie. Does not influence non equilibrium calculations.
        kwargs : dict
            Kwargs to pass to defect_finder. 'verbose' is set to True by default.

        Returns
        -------
        entry : DefectEntry
            DefectEntry object

        """
        if not kwargs:
            kwargs = {'verbose':True}
        elif 'verbose' not in kwargs.keys():
            kwargs['verbose'] = True
        defect = defect_finder(defect_structure, bulk_structure,**kwargs)
        if not defect:
            raise ValueError('Cannot create DefectEntry from empty defect object')
        defect.set_charge(charge)
        defect.set_label(label)
        if multiplicity:
            defect.set_multiplicity(multiplicity)
        else:
            try:
                new_multiplicity = defect.get_multiplicity()
                defect.set_multiplicity(new_multiplicity)
            except NotImplementedError:
                warnings.warn(f'get_multiplicity not implemented for {defect.defect_type}, setting multiplicity to 1')
                defect.set_multiplicity(1)
        
        return DefectEntry(defect, energy_diff, corrections, data)


    @staticmethod
    def from_vasp_directories(
                            path_defect,
                            computed_entry_bulk=None,
                            path_bulk=None,
                            corrections={},
                            multiplicity=1,
                            data=None,
                            label=None,
                            initial_structure=False,
                            function=None,
                            function_kwargs={},
                            computed_entry_kwargs={},
                            finder_kwargs={}):
        """
        Generate DefectEntry object from VASP directories read with Pymatgen.

        Parameters
        ----------
        path_defect : str
            Path of VASP defect calculation.
        computed_entry_bulk : VaspJob
            ComputedStructureEntry of the bulk calculation.
        path_bulk : str
            If `computed_entry_bulk` is not provided, read directly from VASP directory. 
        corrections : dict
            Dict of corrections for defect formation energy. All values will be summed and
            added to the defect formation energy.
        multiplicity : int
            Multiplicity of defect within the supercell. 
            If set to None is attempted to be determined automatically with Pymatgen. The default is 1.
        data : dict
            Store additional data in dict format.
        label : str
            Additional label to add to defect specie. Does not influence non equilibrium calculations.
        initial_structure : bool
            Use initial structure for defect recognition. Useful when relaxations are large and 
            defect_finder struggles to find the right defects.
        function : function
            Function to apply to DefectEntry. Useful to automate custom entry modification.
            The function can modify entry attributes and returns None.
        function_kwargs : dict
            Kwargs to pass to custom function.
        computed_entr_kwargs : dict
            Kwargs to pass to Vasprun.get_computed_entry. 
        finder_kwargs : dict
            Kwargs to pass to `defect_finder`.

        Returns
        -------
        entry : DefectEntry
            DefectEntry object

        """ 
        from pymatgen.io.vasp.outputs import Vasprun

        if initial_structure:
            if computed_entry_kwargs:
                if 'data' in computed_entry_kwargs.keys():
                    computed_entry_kwargs['data'].append('ionic_steps')
                else:
                    computed_entry_kwargs['data'] = ['ionic_steps']
            else:
                computed_entry_kwargs = {'data':['ionic_steps']}
        computed_entry_defect = _get_computed_entry_from_path(path_defect,**computed_entry_kwargs)
        if not computed_entry_bulk:
            if path_bulk:
                vasprun_bulk = os.path.join(path_bulk,'vasprun.xml')
                computed_entry_bulk = Vasprun(vasprun_bulk,parse_dos=False,parse_eigen=False,parse_potcar_file=False).get_computed_entry(**computed_entry_kwargs)
            else:
                raise ValueError('Either bulk ComputedEntry or path of bulk calculation has to be provided')

        entry = DefectEntry.from_computed_entries(
                                                computed_entry_defect=computed_entry_defect,
                                                computed_entry_bulk=computed_entry_bulk,
                                                corrections=corrections,
                                                multiplicity=multiplicity,
                                                data=data,
                                                label=label,
                                                initial_structure=initial_structure,
                                                **finder_kwargs)
        
        if function:
            function(entry,**function_kwargs)
        return entry
        

    def defect_concentration(self,
                            vbm=0,
                            chemical_potentials=None,
                            temperature=300,
                            fermi_level=0.0, 
                            per_unit_volume=True,
                            eform_kwargs={},
                            **kwargs):
        """
        Compute the defect concentration.
        If `defect_concentration_function` is set, the custom function is called, 
        otherwise, the concentration is computed in the dilute limit as:

        n = N * (1 / e^(Ef/kT) + 1), where:

        - N is the site multiplicity (in cm^-3 or per unit cell)
        - Ef is the formation energy of the defect.
        - k is the Boltzmann constant.
        - T is the temperature.

        A custom function can be set when initializing the `DefectEntry` or by using the
        'set_defect_concentration_function` method. Use `reset_defect_concentration_function`
        to restore the default behaviour.
        If a custom function is given, the input must have the same args as this function,
        with the possibility to add more kwargs. 

        Parameters
        ----------
        vbm : float
            Valence band maximum of bulk calculation in eV
        chemical_potentials : dict
            Chemical potentials of the elements involved in the defect.
        temperature : float
            Temperature in Kelvin.
        fermi_level : float
            Fermi level in eV (with respect to the VBM).
        per_unit_volume : bool
            Compute concentrations per unit volume using `self.defect.bulk_volume`.
        eform_kwargs : dict
            Kwargs to pass to `self.formation_energy`.
        kwargs : dict
            Kwargs to pass to custom function.
                
        Returns:
        --------
        conc : float
            Defect concentration in cm^-3 or per unit cell.

        """
        if self.defect_concentration_function:
            return self.defect_concentration_function(self,
                                                    vbm=vbm,
                                                    chemical_potentials=chemical_potentials,
                                                    temperature=temperature,
                                                    fermi_level=fermi_level,
                                                    per_unit_volume=per_unit_volume,
                                                    eform_kwargs=eform_kwargs,
                                                    **kwargs)
        
        
        else:
            return self._default_defect_concentration(vbm=vbm,
                                                    chemical_potentials=chemical_potentials,
                                                    temperature=temperature,
                                                    fermi_level=fermi_level,
                                                    per_unit_volume=per_unit_volume,
                                                    eform_kwargs=eform_kwargs)


    def _default_defect_concentration(self,vbm,chemical_potentials,temperature,
                                      fermi_level,per_unit_volume,eform_kwargs):
        """Default function for the defect concentration calculation"""
        n = self.defect.site_concentration_in_cm3 if per_unit_volume else self.multiplicity 
        eform = self.formation_energy(
                                vbm=vbm,
                                chemical_potentials=chemical_potentials,
                                fermi_level=fermi_level,
                                temperature=temperature,
                                **eform_kwargs)
        
        conc = n * fermi_dirac(eform,temperature) # maxwell_boltzmann(eform,temperature) # 
        return conc


    def formation_energy(self,
                        vbm=0,
                        chemical_potentials=None,
                        fermi_level=0,
                        temperature=0,
                        **kwargs):
        """
        Compute the formation energy of the defect.
        If `formation_energy_function` is set, the custom function is called, 
        otherwise, the formation energy is computed as:

        Ef = E_D - E_B + q(eVBM + fermi_level) + Ecorr - sum_i [(ni(D) - ni(B))* mu_i]

        where:

        - E_D is the energy of the defective cell
        - E_B is the energy of the pristine (bulk) cell
            Note: E_D - E_B is `self.energy_diff`
        - eVBM is the valence band maximum energy
        - fermi_level is the chemical potential of electrons
        - Ecorr are correction terms
        - ni are the number of particles in the defective and pristine cells
        - mu_i are the chemical potentials of the elements.

        A custom function can be set when initializing the `DefectEntry` or by using the
        'set_formation_energy_function` method. Use `reset_formation_energy_function`
        to restore the default behaviour.
        If a custom function is given, the input must have the same args as this function,
        with the possibility to add more kwargs. 

        Parameters
        ----------
        vbm : float
            Valence band maximum of bulk calculation in eV
        chemical_potentials : dict
            Chemical potentials of the elements involved in the defect.
        fermi_level : float
            Fermi level in eV (with respect to the VBM).
        temperature : float
            Temperature in Kelvin. If a custom function is not passed, this parameter has no effect.
        kwargs : dict
            Kwargs to pass to custom function.
        
        Returns
        -------
        formation_energy : float
            Formation energy in eV.

        """
        if self._formation_energy_function:
            return self.formation_energy_function(self,
                                                vbm=vbm,
                                                chemical_potentials=chemical_potentials,
                                                fermi_level=fermi_level,
                                                temperature=temperature,
                                                **kwargs)
            
        else:
            return self._default_formation_energy(
                                            vbm=vbm,
                                            chemical_potentials=chemical_potentials,
                                            fermi_level=fermi_level)
        
    
    def _default_formation_energy(self,vbm,chemical_potentials,fermi_level):
        """Default function for the formation energy calculation"""
        formation_energy = (self.energy_diff + self.charge*(vbm+fermi_level) + 
                       sum([ self.corrections[correction_type]  for correction_type in self.corrections ]) 
                        ) 
        chempot_term = 0
        if chemical_potentials:
            chempot_term = -1 * sum([self.delta_atoms[el]*chemical_potentials[el] for el in self.delta_atoms])
  
        formation_energy = formation_energy + chempot_term    
        return formation_energy
    
    
    def relaxation_volume(self,stress_bulk,bulk_modulus,stress_defect=None,corrections={}):
        """
        Calculate relaxation volume from stresses.
        Stress of defect calculation needs to be provided either directly or in data dict with "stress" key.

        Parameters
        ----------
        stress_bulk : np.array
            Stresses from bulk calculation in kbar (units of VASP output)
        bulk_volume : float
            Cell volume of bulk calculation in A°^3.
        bulk_modulus : float
            Bulk modulus in GPa.
        stress_defect : np.array
            Stresses from defect calculation in kbar (units of VASP output)
        corrections : bool
            Add correction terms to the residual stress tensor.
            
        Returns
        -------
        rel_volume : float
            Relaxation volume in A°^3.

        """
        if not stress_defect:
            if 'stress' in self.data.keys():
                stress_defect = self.data['stress']
            else:
                raise ValueError('Stress of defect calculation needs to be provided, either directly or in data dict with "stress" key.')
        
        return get_relaxation_volume(stress_defect=stress_defect,
                                    stress_bulk=stress_bulk,
                                    bulk_volume=self.defect.bulk_volume,
                                    bulk_modulus=bulk_modulus,
                                    corrections=corrections)





def _get_computed_entry_from_path(path,**kwargs):
    """
    Get Pymatgen's ComputedEntry object from VASP calculation path with additional parameters for defect entries.
    Pass kwargs to Vasprun.get_computed_entry.
    """
    from pymatgen.io.vasp.outputs import Vasprun
    vasprun_defect = os.path.join(path,'vasprun.xml')
    if kwargs:
        if 'data' in kwargs.keys(): # computed entry kwarg and list member are called the same
            if 'parameters' not in kwargs['data']:
                kwargs['data'].append('parameters')
        else:
            kwargs['data'] = ['parameters']
    else:
        kwargs = {'data':['parameters']}     
    return Vasprun(vasprun_defect,parse_dos=False,parse_eigen=False,parse_potcar_file=False).get_computed_entry(**kwargs)


def fermi_dirac(E,T):
    """
    Returns the defect occupation as a function of the formation energy,
    using the Fermi-Dirac distribution with chemical potential equal to 0. 

    Parameters
    ----------
    E : float
        Energy in eV
    T : float
        Temperature in Kelvin.
    """
    from scipy.special import expit
    exponent = E/(kb*T)
    return expit(-exponent)


def maxwell_boltzmann(E, T, clip=True):
    """
    Maxwell-Boltzmann distribution function.
    Can be clipped to prevent divergence.

    Parameters
    ----------
    E : float
        Energy in eV.
    fermi : float
        Fermi level in eV.
    T : float
        Temperature in kelvin.
    clip : bool
        Clip exponential factor to 700 to avoid divergence, returns 1e304.

    Returns
    -------
    occupation : float
        Maxwell-Boltzmann occupation probability at energy E.

    """
    factor = -E / (kb*T)
    if clip:
        factor = np.clip(factor,None,700)
    return np.exp(factor)







import os.path as op
import importlib

from pymatgen.io.vasp.outputs import Locpot
from pymatgen.analysis.defects.corrections.freysoldt import get_freysoldt_correction, plot_plnr_avg

from ..structure import defect_finder


def get_freysoldt_correction_from_locpot(
                                    charge,
                                    dielectric_constant,
                                    defect_path_locpot,
                                    bulk_path_locpot,
                                    get_correction_data=True,
                                    get_plot=True,
                                    plot_axis_index=0,
                                    finder_kwargs={},
                                    **kwargs):
    """
    Get Freysoldt correction from defect and bulk LOCPOT files, using `pymatgen-analysis-defects`.

    Parameters
    ----------
    charge : int
        Charge of the defect
    dielectric_constant : float
        Dielectric constant of the material
    defect_path_locpot : str
        Defect LOCPOT file.
    bulk_path_locpot : str
        Bulk LOCPOT file.
    get_correction_data : bool
        Return pymatgen's CorrectionResult object. If False only the correction value is returned.
    get_plot : bool
        Get Matplotlib object with plot. The default is False.
    plot_axis_index : int
        Axis for planar average plot. The default is 0.
    finder_kwargs : dict
        Kwargs to pass to defect_finder.
    kwargs : dict
        Kwargs to pass to pymatgen's get_freysoldt_correction.

    Returns
    -------
    correction, ax : tuple
        CorrectionResult object if get_correction_data is set to True, else just the float with correction value, matplotlib axis object
    """
    q = charge
    defect_locpot = Locpot.from_file(defect_path_locpot)
    bulk_locpot = Locpot.from_file(bulk_path_locpot)
    if finder_kwargs:
        if 'verbose' not in finder_kwargs.keys():
            finder_kwargs['verbose'] = True
    defect = defect_finder(defect_locpot.structure,bulk_locpot.structure,**finder_kwargs)
    defect_frac_coords = defect.site.frac_coords
    lattice = defect_locpot.structure.lattice
    correction = get_freysoldt_correction(q=q,dielectric=dielectric_constant,defect_locpot=defect_locpot,
                                    bulk_locpot=bulk_locpot,defect_frac_coords=defect_frac_coords,
                                    lattice=lattice,**kwargs)
    
    corr = correction if get_correction_data else correction.correction_energy
    if get_plot:
        ax = plot_plnr_avg(correction.metadata['plot_data'][plot_axis_index])
        return corr, ax
    else:
        return corr 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 14 14:48:55 2021

@author: villa
"""
import warnings
from monty.json import MSONable
from monty.json import jsanitize
import numpy as np

from .analysis import DefectConcentrations, SingleDefConc
from .defects import get_defect_from_string
import copy
import os.path as op
import json
import os


class Conductivity:
    """
    Class that handles conductivity calculations.
    """
    def __init__(self,mobilities):
        """
        Parameters
        ----------
        mobilities : dict
            Dictionary with mobility values for the defect species in [cm^2/(V*s)].
            Keys must contain "electrons", "holes" and the defect species names.
        """
        self.mobilities = mobilities
        
        
    def get_conductivity(self,carrier_concentrations,defect_concentrations):
        """
        Calculate conductivity from the concentrations of electrons, holes and defects and their mobilities.
        
        Parameters
        ----------
        carrier_concentrations : list
            List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
        defect_concentrations : list
            Defect concentrations in the same format as the output of DefectsAnalysis in cm^-3. 

        Returns
        -------
        sigma : float
            Conductivity in S/m.

        """
        e = 1.60217662e-19
        mob = self.mobilities
        cc = carrier_concentrations
        sigma_el = e * (mob['holes']*cc[0]+ mob['electrons']*cc[1])
        dc = defect_concentrations
        sigma_ionic = 0
        for d in dc:
            dname = d['name']
            if dname in mob.keys():
                sigma_ionic += mob[dname] * d['conc'] * abs(d['charge']) * e
        sigma = sigma_el + sigma_ionic * 1e02 # conversion from S/cm to S/m
        
        return sigma


    def get_conductivities_from_thermodata(self,thermodata):
        """
        Calculate conductivities from defect and carrier concentrations in
        thermodynamic data. Only the defect species which are present in the
        mobility dict are considered for the calculation of the conductivity.

        Parameters
        ----------
        thermodata: ThermoData
            ThermoData object that contains the thermodynamic data.

        Returns
        -------
        conductivities: list
            List with conductivity values in S/m.

        """
        conductivities = []
        for i in range(0,len(thermodata.carrier_concentrations)):
            carrier_concentrations = thermodata.carrier_concentrations[i]
            dc = thermodata.defect_concentrations[i] 
            defect_concentrations = []
            for d in dc:
                if d['name'] in self.mobilities.keys():
                    defect_concentrations.append(d)
            conductivity = self.get_conductivity(
                                                carrier_concentrations=carrier_concentrations,
                                                defect_concentrations=defect_concentrations)
            conductivities.append(conductivity)
        
        return conductivities 



class DefectThermodynamics:
    """
    Class that handles the calculations of defect equilibriua under different conditions.
    """
    
    def __init__(self,
                defects_analysis,
                bulk_dos,
                fixed_concentrations=None,
                external_defects=[],
                xtol=1e-05,
                eform_kwargs={},
                dconc_kwargs={}):
        """
        Parameters
        ----------
        defects_analysis : DefectsAnalysis
            DefectsAnalysis object.
        bulk_dos : dict or Dos
            Density of states to integrate. Can be provided as density of states D(E)
            or using effective masses.

            Format for effective masses is a dict with the following keys:

            - "m_eff_h" : holes effective mass in units of m_e (electron mass)
            - "m_eff_e" : electrons effective mass in units of m_h
            - `band_gap` : needs to be provided in arguments

            Format for explicit DOS (dictionary) with the following keys:

            - 'energies' : list or np.array with energy values
            - 'densities' : list or np.array with total density values
            - 'structure' : pymatgen Structure of the material, needed for DOS volume and charge normalization

            Alternatively, a pymatgen Dos object (Dos, CompleteDos, or FermiDos).

        fixed_concentrations: dict
            Dictionary with fixed concentrations. Keys are defect entry names in the standard
            format, values are the concentrations. (ex {'Vac_Na':1e20}) 
        external_defects : list
            List of external defect concentrations (not present in defect entries).
            Must either be a list of dictionaries with {'charge': float, 'conc': float} 
            or a list of SingleDefConc objects. 
        xtol : float
            Tolerance for `bisect` (scipy) to solve charge neutrality.
        eform_kwargs : dict
            Kwargs to pass to `entry.formation_energy`.
        dconc_kwargs : dict
            Kwargs to pass to `entry.defect_concentration`.
        """
        self.da = defects_analysis
        self.bulk_dos = bulk_dos
        self.fixed_concentrations = fixed_concentrations if fixed_concentrations else None
        self.external_defects = external_defects or []
        self.xtol = xtol
        self.eform_kwargs = eform_kwargs
        self.dconc_kwargs = dconc_kwargs


    def get_pO2_thermodata(
                        self,
                        reservoirs,
                        temperature=None,
                        name=None):
        """
        Calculate defect and carrier concentrations as a function of the oxygen partial pressure.

        Parameters
        ----------
        reservoirs : dict, Reservoirs or PressureReservoirs
            Object with partial pressure values as keys and chempots dictionary as values.
        temperature : float
            Temperature in Kelvin. If None reservoirs.temperature is used.
        name : str
            Name to assign to ThermoData.

        Returns
        -------
        thermodata : ThermoData
            ThermoData object that contains the thermodynamic data:

            partial_pressures : (list)
                List of partial pressure values in atm.
            defect_concentrations : (list)
                List of DefectConcentrations objects (cm^-3)
            carrier_concentrations : (list)
                List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
            fermi_levels : (list)
                list of Fermi level values in eV.

        """
        res = reservoirs
        if temperature:
            T = temperature
        elif hasattr(res,'temperature'):
            T = res.temperature
        else:
            raise ValueError('Temperature needs to be provided or to be present ad attribute in PressureReservoirs object')
        partial_pressures = list(res.keys())
        defect_concentrations = []
        carrier_concentrations = []
        fermi_levels=[]

        for r,mu in res.items():
            single_thermodata = self.get_single_point_thermodata(
                                                            chemical_potentials=mu, 
                                                            temperature=T
                                                            )

            defect_concentrations.append(single_thermodata['defect_concentrations'])
            carrier_concentrations.append(single_thermodata['carrier_concentrations'])
            fermi_levels.append(single_thermodata['fermi_levels'])
        
        thermodata = {}
        thermodata['partial_pressures'] = partial_pressures 
        thermodata['defect_concentrations'] = defect_concentrations 
        thermodata['carrier_concentrations'] = carrier_concentrations
        thermodata['fermi_levels'] = fermi_levels
        
        thermodata = ThermoData(thermodata,temperature=temperature,name=name)
        
        return thermodata


    def get_pO2_quenched_thermodata(
                                self,
                                reservoirs,
                                initial_temperature,
                                final_temperature,
                                quenched_species=None,
                                quench_elements=False,
                                name=None):
        """
        Calculate defect and carrier concentrations as a function of oxygen partial pressure with quenched defects. 
        It is possible to select which defect species to quench and which ones are free to equilibrate.
        Frozen defects in the inputs of the class are still considered.

        Parameters
        ----------
        reservoirs : dict, Reservoirs or PressureReservoirs
            Object with partial pressure values as keys and chempots dictionary as values.
        initial_temperature : float
            Value of initial temperature in K.
        final_temperature : float
            Value of final temperature in K.
        quenched_species : list
            List of defect species to quench. If None all defect species are quenched.The default is None.
        quench_elements : bool
            If True the total concentrations of elements at high temperature go in the charge neutrality at low temperature.
            If False the quenched concentrations are the ones of single defect species (e.g. elements are not allowed
            to equilibrate on different sites). The default is False.
        name : str
            Name to assign to ThermoData.
            
        Returns
        -------
        thermodata : ThermoData
            ThermoData object that contains the thermodynamic data:

            partial_pressures : (list)
                List of partial pressure values in atm.
            defect_concentrations : (list)
                List of DefectConcentrations objects (cm^-3)
            carrier_concentrations : (list)
                List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
            fermi_levels : (list)
                list of Fermi level values in eV.


        """        
        res = reservoirs
        if hasattr(res,'temperature'):
            if res.temperature != initial_temperature:
                warnings.warn('PressureReservoirs temperature is not set to the initial quenching temperature',UserWarning)
        partial_pressures = list(res.keys())
        fermi_levels = []
        defect_concentrations = []
        carrier_concentrations = []
                
        for r,mu in res.items():
            single_quenched_thermodata = self.get_single_point_quenched_thermodata(
                                                        chemical_potentials=mu,
                                                        initial_temperature=initial_temperature,
                                                        final_temperature=final_temperature,
                                                        quenched_species=quenched_species,
                                                        quench_elements=quench_elements)
            
            carrier_concentrations.append(single_quenched_thermodata['carrier_concentrations'])
            defect_concentrations.append(single_quenched_thermodata['defect_concentrations'])
            fermi_levels.append(single_quenched_thermodata['fermi_levels'])
 
        thermodata = {}
        thermodata['partial_pressures'] = partial_pressures
        thermodata['fermi_levels'] = fermi_levels
        thermodata['defect_concentrations'] = defect_concentrations
        thermodata['carrier_concentrations'] = carrier_concentrations
        
        thermodata = ThermoData(thermodata,temperature=(initial_temperature,final_temperature),name=name)
        
        return thermodata


    def get_single_point_thermodata(self,
                                    chemical_potentials,
                                    temperature,
                                    fixed_concentrations=None,
                                    external_defects=None,
                                    name=None,
                                    eform_kwargs=None,
                                    dconc_kwargs=None):
        """
        Compute carrier concentrations, defect concentrations and Fermi level for 
        a single set of chemical potentials.

        Parameters
        ----------
        chemical_potentials : dict or Chempots
            Chempots object containing chemical potentials.
        temperature : int
            Temperature in K.
        fixed_concentrations: dict
            Dictionary with fixed concentrations. Keys are defect entry names in the standard
            format, values are the concentrations. (ex {'Vac_Na':1e20}) 
        external_defects : list
            List of external defect concentrations (not present in defect entries).
            Must either be a list of dictionaries with {'charge': float, 'conc': float} 
            or a list of SingleDefConc objects. 
        name : str
            Label for ThermoData.
        eform_kwargs : dict
            Kwargs to pass to `entry.formation_energy`.
        dconc_kwargs : dict
            Kwargs to pass to `entry.defect_concentration`.

        Returns
        -------
        thermodata : ThermoData
            ThermoData object that contains the thermodynamic data:

            defect_concentrations : (DefectConcentrations)
                DefectConcentrations object (cm^-3).
            carrier_concentrations : (tuple)
                Tuple with intrinsic carriers concentrations in cm^-3 (holes,electrons).
            fermi_levels : (float)
                Fermi level value in eV.

        """
        dos = self.bulk_dos
        fixed_df = fixed_concentrations or self.fixed_concentrations
        ext_df = external_defects or self.external_defects
        eform_kwargs = eform_kwargs if eform_kwargs is not None else self.eform_kwargs
        dconc_kwargs = dconc_kwargs if dconc_kwargs is not None else self.dconc_kwargs

        fermi_level = self.da.solve_fermi_level(
                                                chemical_potentials=chemical_potentials,
                                                bulk_dos=dos,
                                                temperature=temperature,
                                                fixed_concentrations=fixed_df,
                                                external_defects=ext_df,
                                                xtol=self.xtol,
                                                eform_kwargs=eform_kwargs,
                                                dconc_kwargs=dconc_kwargs)
        
        carrier_concentrations = self.da.carrier_concentrations(
                                                            bulk_dos=dos,
                                                            temperature=temperature,
                                                            fermi_level=fermi_level)
        
        defect_concentrations = self.da.defect_concentrations(
                                                            chemical_potentials=chemical_potentials,
                                                            temperature=temperature,
                                                            fermi_level=fermi_level,
                                                            fixed_concentrations=fixed_df,
                                                            eform_kwargs=eform_kwargs,
                                                            **dconc_kwargs)
        if ext_df:
            for df in ext_df:         
                defect_concentrations.append(df)
        
        thermodata = {'carrier_concentrations':carrier_concentrations,
                      'defect_concentrations':defect_concentrations,
                      'fermi_levels':fermi_level}
        
        return ThermoData(thermodata,temperature=temperature,name=name)
    

    def get_single_point_quenched_thermodata(
                                            self,
                                            chemical_potentials,
                                            initial_temperature,
                                            final_temperature,
                                            quenched_species=None,
                                            quench_elements=False,
                                            fixed_concentrations=None,
                                            external_defects=None,
                                            name=None):
        """
        Compute carrier concentrations, defect concentrations and Fermi level for 
        a single set of chemical potentials.

        Parameters
        ----------
        chemical_potentials : dict or Chempots
            Chempots object containing chemical potentials.
        initial_temperature : float
            Initial temperature in K.
        final_temperature : float
            Final temperature in K
        quenched_species : list
            List of defect species to quench. If None all defect species are quenched.
        quench_elements : bool
            If True the total concentrations of elements at high temperature go in the charge neutrality at low temperature.
            If False the quenched concentrations are the ones of single defect species (e.g. elements are not allowed
            to equilibrate on different sites).
        fixed_concentrations : dict
            Dictionary with fixed concentrations. Keys are defect entry names in the standard
            format, values are the concentrations. (ex {'Vac_Na':1e20}) 
        external_defects : list
            List of external defect concentrations (not present in defect entries).
            Must either be a list of dictionaries with {'charge': float, 'conc': float} 
            or a list of SingleDefConc objects. 
        name : str
            Label for ThermoData.

        Returns
        -------
        thermodata : ThermoData
            ThermoData object that contains the thermodynamic data:

            defect_concentrations : (DefectConcentrations)
                DefectConcentrations object (cm^-3).
            carrier_concentrations : (tuple)
                Tuple with intrinsic carriers concentrations in cm^-3 (holes,electrons).
            fermi_levels : (float)
                Fermi level value in eV.

        """
        fixed_df = fixed_concentrations or self.fixed_concentrations
        ext_df = external_defects or self.external_defects
        
        single_thermodata = self.get_single_point_thermodata(
                                                    chemical_potentials=chemical_potentials,
                                                    temperature=initial_temperature,
                                                    fixed_concentrations=fixed_df,
                                                    external_defects=ext_df
                                                    )

        if quench_elements:
            c1 = single_thermodata['defect_concentrations'].elemental
        else:
            c1 = single_thermodata['defect_concentrations'].total

        if quenched_species is None:
            quenched_concentrations = c1.copy()
        else:
            if fixed_df:
                quenched_concentrations = copy.deepcopy(fixed_df)
            else:
                quenched_concentrations = {}
            for k in quenched_species:
                quenched_concentrations[k] = c1[k]
        
        single_quenched_thermodata = self.get_single_point_thermodata(
                                                    chemical_potentials=chemical_potentials,
                                                    temperature=final_temperature,
                                                    fixed_concentrations=quenched_concentrations,
                                                    external_defects=ext_df
                                                    )    
            
        return single_quenched_thermodata


    def _update_variable_species_concentration(self,variable_defect_specie, c, fixed_df, ext_df):

        if type(variable_defect_specie) == str :
            fixed_df.update({variable_defect_specie:c})
            variable_defect_specie_str = variable_defect_specie
            return fixed_df, ext_df, variable_defect_specie_str 
        elif type(variable_defect_specie) == dict:
            vds = variable_defect_specie
            try:
                variable_defect_specie_str = get_defect_from_string(vds['name']).specie
            except:
                variable_defect_specie_str = vds['name']
            
            if not ext_df:
                single_df_conc = SingleDefConc(
                                name=vds['name'],
                                charge=vds['charge'],
                                conc=c) # variable concentration
                ext_df_update = DefectConcentrations([single_df_conc])
            else:
                if type(ext_df) == dict:
                    ext_df = DefectConcentrations(ext_df)
                ext_df_update = ext_df.copy()
                if vds['name'] not in ext_df_update.names:
                    single_df_conc = SingleDefConc(
                                    name=vds['name'],
                                    charge=vds['charge'],
                                    conc=c)
                    ext_df_update.append(single_df_conc)

                single_df_conc = ext_df_update.select_concentrations(name=vds['name'],charge=vds['charge'])[0]
                single_df_conc.conc = c 

            return fixed_df, ext_df_update, variable_defect_specie_str  

        else:
            raise ValueError('Variable species has to be either a string (if present in DefectsAnalysis) or dict (if not present in DefectsAnalysis)')

                 
    def get_variable_species_thermodata(
                                        self,
                                        variable_defect_specie,
                                        concentration_range,
                                        chemical_potentials,
                                        temperature,
                                        external_defects=[],
                                        npoints=50,
                                        name=None):
        """
        Calculate defect and carrier concentrations as a function of the concentration of a particular 
        defect species (usually a dopant).

        Parameters
        ----------
        variable_defect_specie : str or dict

            Variable species. Possible formats are:

            - str: name or element, if the variable defect species is in defect entries.
            - dict : {'name':str,'charge':int or float} if the variable species is not in defect entries. 

        concentration_range : tuple or list
            Range of the concentration of the variable species in cm^-3.
        chemical_potentials : dict or Chempots
            Chempots object containing chemical potentials.
        temperature : float
            Temperature.
        npoints : int
            Number of points to divide concentration range.
        external_defects : list
            List of external defect concentrations (not present in defect entries).
            Must either be a list of dictionaries with {'charge': float, 'conc': float} 
            or a list of SingleDefConc objects. 
        name : str
            Label for ThermoData.

        Returns
        -------
        thermodata : ThermoData

            ThermoData object that contains the thermodynamic data:

            variable_defect_specie : (str)
                Name of variable defect species.
            variable_concentrations : (list)
                List of concentrations of variable species. 
            defect_concentrations : (list)
                List of DefectConcentrations objects in the same format as the output of DefectsAnalysis. 
            carrier_concentrations : (list)
                List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
            fermi_levels : (list)
                List of Fermi level values in eV.

        """
        carrier_concentrations = []
        defect_concentrations = []
        fermi_levels = []
        
        concentrations = np.logspace(start=np.log10(concentration_range[0]),stop=np.log10(concentration_range[1]),num=npoints)
        fixed_df = self.fixed_concentrations.copy() if self.fixed_concentrations else {}
        ext_df = external_defects or self.external_defects
        if ext_df and type(ext_df) != DefectConcentrations:                
            ext_df = DefectConcentrations(ext_df)

        for c in concentrations:
            fixed_df, ext_df, variable_defect_specie_str = self._update_variable_species_concentration(
                                                    variable_defect_specie,c,fixed_df,ext_df)
            
            single_thermodata = self.get_single_point_thermodata(
                                            chemical_potentials=chemical_potentials,
                                            temperature=temperature,fixed_concentrations=fixed_df,
                                            external_defects=ext_df
                                            )
            
            defect_concentrations.append(single_thermodata['defect_concentrations'])
            carrier_concentrations.append(single_thermodata['carrier_concentrations'])
            fermi_levels.append(single_thermodata['fermi_levels'])
            
        data = {}
        data['variable_defect_specie'] = variable_defect_specie_str
        data['variable_concentrations'] = concentrations
        data['defect_concentrations'] = defect_concentrations
        data['carrier_concentrations'] = carrier_concentrations
        data['fermi_levels'] = fermi_levels
        
        thermodata = ThermoData(data,temperature=temperature,name=name)
        
        return thermodata


    def get_variable_species_quenched_thermodata(
                                                self,
                                                variable_defect_specie,
                                                concentration_range,
                                                chemical_potentials,
                                                initial_temperature,
                                                final_temperature,
                                                quenched_species=None,
                                                quench_elements=False,
                                                external_defects=[],
                                                npoints=50,
                                                name=None):
        """
        Calculate defect and carrier concentrations as a function of the concentration of a particular 
        defect species (usually a dopant) with quenched defects.
        It is possible to select which defect species to quench and which ones are free to equilibrate.
        Frozen defects in the inputs of the class are still considered.

        Parameters
        ----------
        variable_defect_specie : str or dict

            Variable species. Possible formats are:

            - str: name or element, if the variable defect species is in defect entries.
            - dict : {'name':str,'charge':int or float} if the variable species is not in defect entries. 
        
        concentration_range : tuple or list
            Range of the concentration of the variable species in cm^-3.
        chemical_potentials : dict or Chempots
            Chempots object containing chemical potentials.
        initial_temperature : float
            Value of initial temperature in K.
        final_temperature : float
            Value of final temperature in K.
        quenched_species : list
            List of defect species to quench. If None all defect species are quenched.
        quench_elements : bool
            If True the total concentrations of elements at high temperature go in the charge neutrality at low temperature.
            If False the quenched concentrations are the ones of single defect species (e.g. elements are not allowed
            to equilibrate on different sites).
        external_defects : list
            List of external defect concentrations (not present in defect entries).
            Must either be a list of dictionaries with {'charge': float, 'conc': float} 
            or a list of SingleDefConc objects. 
        npoints : int
            Number of points to divide concentration range.
        name : str
            Name to assign to ThermoData.

        Returns
        -------
        thermodata : ThermoData

            ThermoData object that contains the thermodynamic data:

            variable_defect_specie : (str)
                Name of variable defect species.
            variable_concentrations : (list)
                List of concentrations of variable species. 
            defect_concentrations : (list)
                List of DefectConcentrations objects in the same format as the output of DefectsAnalysis. 
            carrier_concentrations : (list)
                List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
            fermi_levels : (list)
                List of Fermi level values in eV.

        """
        carrier_concentrations = []
        defect_concentrations = []
        fermi_levels = []
        
        concentrations = np.logspace(start=np.log10(concentration_range[0]),stop=np.log10(concentration_range[1]),num=npoints)
        fixed_df = self.fixed_concentrations.copy() if self.fixed_concentrations else {}
        ext_df = external_defects or self.external_defects
        for c in concentrations:
            fixed_df, ext_df, variable_defect_specie_str = self._update_variable_species_concentration(
                                                    variable_defect_specie,c,fixed_df,ext_df)
                        
            single_quenched_thermodata = self.get_single_point_quenched_thermodata(
                                                    chemical_potentials=chemical_potentials,
                                                    initial_temperature=initial_temperature,
                                                    final_temperature=final_temperature,
                                                    quenched_species=quenched_species,
                                                    quench_elements=quench_elements,
                                                    fixed_concentrations=fixed_df,
                                                    external_defects=ext_df
                                                    )
            
            defect_concentrations.append(single_quenched_thermodata['defect_concentrations'])
            carrier_concentrations.append(single_quenched_thermodata['carrier_concentrations'])
            fermi_levels.append(single_quenched_thermodata['fermi_levels'])
                
            
        data = {}
        data['variable_defect_specie'] = variable_defect_specie_str
        data['variable_concentrations'] = concentrations
        data['fermi_levels'] = fermi_levels
        data['defect_concentrations'] = defect_concentrations
        data['carrier_concentrations'] = carrier_concentrations
        
        thermodata = ThermoData(data,temperature=(initial_temperature,final_temperature),name=name)
        
        return thermodata



class ThermoData(MSONable):
    
    "Class to handle defect thermodynamics data"
    
    def __init__(self,thermodata,temperature=None,name=None):
        """
        Class that handles dict of defect thermodynamics data.
        
        Parameters
        ----------
        thermodata : dict
            Dict that contains the thermodynamic data, typically output from 
            the methods in the `DefectThermodynamics` class.
            Keys of the dict are set as attributes of ThermoData.

            The possible items are:

            partial_pressures : (list)
                List of partial pressure values.
            variable_defect_specie : (str)
                Name of variable defect species.
            variable_concentrations : (list)
                List of concentrations of variable species. 
            defect_concentrations : (list)
                List of DefectConcentrations objects (cm^-3).
            carrier_concentrations : (list)
                List of tuples with intrinsic carriers concentrations (holes,electrons) in cm^-3.
            conductivities : (list)
                List of conductivity values (in S/m).
            fermi_levels : (list)
                list of Fermi level values in eV.

        temperature : float
            Temperature in K at which the data is computed.
        name : str
            Name to assign to ThermoData.
        """
        self.data = thermodata  
        self.temperature = temperature if temperature else None
        self.name = name if name else None
        for k,v in thermodata.items():
            setattr(self, k, v)
        
    
    def __getitem__(self,key):
        return self.data[key]
    
    def __iter__(self):
        return self.data.__iter__()
    
    def keys(self):
        return self.data.keys()
    
    def values(self):
        return self.data.values()
    
    def items(self):
        return self.data.items()
    
    
    def as_dict(self):
        """
        Returns:
            Json-serializable dict representation of ThermoData
        """
        d = {"@module": self.__class__.__module__,
              "@class": self.__class__.__name__,
              "thermodata":self.data.copy(),    
              "temperature": self.temperature,
              "name": self.name
              }
        if 'defect_concentrations' in self.data.keys():
            d['thermodata']['defect_concentrations'] = [dc.as_dict() for dc in self.defect_concentrations]
        d = jsanitize(d) #convert numpy.float64 to float
        return d        

    def to_json(self,path=None):
        """
        Save ThermoData object as json string or file

        Parameters
        ----------
        path : str
            Path to the destination file.  If None the name of ThermoData is used as filename.

        Returns
        -------
        d : str
            If path is not set a string is returned.
        """
        if not path:
            path = op.join(os.getcwd(),f'thermodata_{self.name}.json')
        d = self.as_dict()
        with open(path,'w') as file:
            json.dump(d,file)
        return
 

    @classmethod
    def from_dict(cls,d):
        """
        Reconstitute a ThermoData object from a dict representation created using
        as_dict().

        Parameters
        ----------
        d : dict 
            Dictionary representation of ThermoData.

        Returns
        -------
        ThermoData object.

        """
        if 'thermodata' in d.keys():
            data = d['thermodata'].copy()
            if 'defect_concentrations' in data.keys():
                data['defect_concentrations'] = [DefectConcentrations.from_dict(dc) for dc in data['defect_concentrations']]
            temperature = d['temperature']
            name = d['name']
        else:
            data = d # recover old 
            temperature = None
            name = None
        return cls(data,temperature,name)
     
         
    @staticmethod
    def from_json(path_or_string):
        """
        Build ThermoData object from json file or string.

        Parameters
        ----------
        path_or_string : str
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        ThermoData object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return ThermoData.from_dict(d)
            
    
    def get_specific_pressures(self,p_values):
        """
        Get ThermoData object only for specific pressure values. The closest pressure value
        present in list is chosen for each provided value.

        Parameters
        ----------
        p_values : list
            List of partial pressure values.

        Returns
        -------
        ThermoData object
        """
        seldata = {}
        for p in p_values:
            pressures = self.partial_pressures
            p_sel = min(pressures, key=lambda x:abs(x-p))
            index = pressures.index(p_sel)
            for k,v in self.data.items():
                for e in v:
                    if v.index(e) == index:
                        if k not in seldata.keys():
                            seldata[k] = []
                        seldata[k].append(e)
        
        temperature = self.temperature if self.temperature else None
        

        name = self.name + 'p_' + '-'.join([str(p) for p in p_values]) if self.name else None
            
        return ThermoData(seldata,temperature=temperature,name=name)
        
    
    def set_data(self,key,value): # chosen over __setitem__ to not accidentally overwrite data
        """
        Set data dictionary.
        """
        self.data[key] = value
        setattr(self, key, value)
        return

